# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import os

import numpy

# -*- coding: UTF-8 -*-
import onnx
from onnx import AttributeProto, GraphProto, TensorProto, helper, numpy_helper  # noqa: F401

if os.path.exists(
    os.path.join(
        os.path.dirname(__file__),
        "..",
        "..",
        "python",
        "tools",
        "symbolic_shape_infer.py",
    )
):
    # Allow running this test script without installing onnxruntime package.
    import sys

    sys.path.append(os.path.join(os.path.dirname(__file__), "..", "..", "python", "tools"))
    from symbolic_shape_infer import SymbolicShapeInference
else:
    from onnxruntime.tools.symbolic_shape_infer import SymbolicShapeInference

import unittest


def unique_element(lst):
    assert len(lst) == 1
    return lst[0]


skipped_models = ["SSD-MobilenetV1", "SSD-int8", "Inception-1-int8"]


class TestSymbolicShapeInference(unittest.TestCase):
    def test_symbolic_shape_infer(self):
        from pathlib import Path

        cwd = os.getcwd()
        test_model_dir = os.path.join(cwd, "..", "models")
        for filename in Path(test_model_dir).rglob("*.onnx"):
            if filename.name.startswith("."):
                continue  # skip some bad model files

            # https://github.com/onnx/models/issues/562
            if any(model_name in str(filename) for model_name in skipped_models):
                print(f"Skip symbolic shape inference on : {filename!s}")
                continue

            print("Running symbolic shape inference on : " + str(filename))
            SymbolicShapeInference.infer_shapes(
                in_mp=onnx.load(str(filename)),
                auto_merge=True,
                int_max=100000,
                guess_output_rank=True,
            )

    def test_mismatched_types(self):
        graph = helper.make_graph(
            [
                helper.make_node(
                    "If",
                    ["x"],
                    ["out"],
                    name="if_node",
                    then_branch=helper.make_graph(
                        [
                            helper.make_node(
                                "Constant",
                                [],
                                ["one_float"],
                                value=helper.make_tensor("one_float_value", TensorProto.FLOAT, [], [1]),
                            )
                        ],
                        "then",
                        [],
                        [helper.make_tensor_value_info("one_float", TensorProto.FLOAT, [])],
                    ),
                    else_branch=helper.make_graph(
                        [
                            helper.make_node(
                                "Constant",
                                [],
                                ["one_double"],
                                value=helper.make_tensor("one_double", TensorProto.DOUBLE, [], [1]),
                            )
                        ],
                        "else",
                        [],
                        [helper.make_tensor_value_info("one_double", TensorProto.DOUBLE, [])],
                    ),
                )
            ],
            "graph",
            [helper.make_tensor_value_info("x", TensorProto.BOOL, [])],
            [helper.make_tensor_value_info("out", TensorProto.FLOAT, [])],
        )
        model = helper.make_model(graph, producer_name="test_mismatched_types")

        with self.assertRaisesRegex(ValueError, r"if_node.*FLOAT.*DOUBLE"):
            SymbolicShapeInference.infer_shapes(model, auto_merge=True)


class TestSymbolicShapeInferenceForOperators(unittest.TestCase):
    def _check_shapes(self, graph, inferred_graph, vis):  # type: (GraphProto, GraphProto, List[ValueInfoProto]) -> None
        names_in_vis = {x.name for x in vis}
        vis = [x for x in graph.value_info if x.name not in names_in_vis] + vis
        inferred_vis = list(inferred_graph.value_info)
        vis = sorted(vis, key=lambda x: x.name)
        inferred_vis = sorted(inferred_vis, key=lambda x: x.name)
        if vis == inferred_vis:
            return
        # otherwise some custom logic to give a nicer diff
        vis_names = {x.name for x in vis}
        inferred_vis_names = {x.name for x in inferred_vis}
        assert vis_names == inferred_vis_names, (vis_names, inferred_vis_names)
        for vi, inferred_vi in zip(vis, inferred_vis, strict=False):
            assert vi == inferred_vi, f"\n{vi}\n{inferred_vi}\n"
        raise AssertionError()

    def test_unsqueeze_opset_11(self):
        graph = helper.make_graph(
            [
                helper.make_node("Unsqueeze", ["input"], ["temp"], axes=[0]),
                helper.make_node("Identity", ["temp"], ["output"]),
            ],
            "Unsqueeze_Test",
            [
                helper.make_tensor_value_info("input", TensorProto.FLOAT, ["b", "s"]),
            ],
            [
                helper.make_tensor_value_info("output", TensorProto.FLOAT, [1, "b", "s"]),
            ],
        )
        model = helper.make_model(graph, producer_name="Unsqueeze_Test_Model")
        model.opset_import[0].version = 11

        inferred = SymbolicShapeInference.infer_shapes(model, auto_merge=True)
        expected_shapes = [
            helper.make_tensor_value_info("temp", TensorProto.FLOAT, [1, "b", "s"]),
            helper.make_tensor_value_info("output", TensorProto.FLOAT, [1, "b", "s"]),
        ]
        self._check_shapes(graph, inferred.graph, expected_shapes)

    def test_unsqueeze_opset_13(self):
        graph = helper.make_graph(
            [
                helper.make_node("Unsqueeze", ["input", "axes"], ["temp"]),
                helper.make_node("Identity", ["temp"], ["output"]),
            ],
            "Unsqueeze_Test",
            [
                helper.make_tensor_value_info("input", TensorProto.FLOAT, ["b", "s"]),
            ],
            [
                helper.make_tensor_value_info("output", TensorProto.FLOAT, ["b", "s", 1]),
            ],
            [
                helper.make_tensor("axes", TensorProto.INT64, [1], [-1]),
            ],
        )
        model = helper.make_model(graph, producer_name="Unsqueeze_Test_Model")
        model.opset_import[0].version = 13

        inferred = SymbolicShapeInference.infer_shapes(model, auto_merge=True)
        expected_shapes = [
            helper.make_tensor_value_info("temp", TensorProto.FLOAT, ["b", "s", 1]),
            helper.make_tensor_value_info("output", TensorProto.FLOAT, ["b", "s", 1]),
        ]
        self._check_shapes(graph, inferred.graph, expected_shapes)

    def test_gather_indices(self):
        graph = helper.make_graph(
            [
                helper.make_node(
                    "Constant",
                    [],
                    ["data"],
                    "constant",
                    value=helper.make_tensor("input", TensorProto.FLOAT, [5], [0.0, 1.0, 2.0, 3.0, 4.0]),
                ),
                helper.make_node("Gather", ["data", "indices"], ["output"], axis=0),
            ],
            "Gather_Test",
            [
                helper.make_tensor_value_info("indices", TensorProto.INT64, ["b"]),
            ],
            [
                helper.make_tensor_value_info("output", TensorProto.FLOAT, ["b"]),
            ],
        )
        model = helper.make_model(graph, producer_name="Gather_Test_Model")
        model.opset_import[0].version = 13

        inferred = SymbolicShapeInference.infer_shapes(model, auto_merge=True)
        expected_shapes = [
            helper.make_tensor_value_info("data", TensorProto.FLOAT, [5]),
            helper.make_tensor_value_info("output", TensorProto.FLOAT, ["b"]),
        ]
        self._check_shapes(graph, inferred.graph, expected_shapes)

    def test_embed_layer_norm(self):
        hidden_size = 32
        initializers = [
            helper.make_tensor(
                "word_embedding",
                TensorProto.FLOAT,
                [100, hidden_size],
                [1.0] * (100 * hidden_size),
            ),
            helper.make_tensor(
                "position_embedding",
                TensorProto.FLOAT,
                [20, hidden_size],
                [1.0] * (20 * hidden_size),
            ),
            helper.make_tensor(
                "segment_embedding",
                TensorProto.FLOAT,
                [2, hidden_size],
                [1.0] * (2 * hidden_size),
            ),
            helper.make_tensor("gamma", TensorProto.FLOAT, [hidden_size], [1.0] * hidden_size),
            helper.make_tensor("beta", TensorProto.FLOAT, [hidden_size], [1.0] * hidden_size),
        ]

        nodes = [
            helper.make_node(
                "EmbedLayerNormalization",
                inputs=[
                    "input_ids",
                    "segment_ids",
                    "word_embedding",
                    "position_embedding",
                    "segment_embedding",
                    "gamma",
                    "beta",
                ],
                outputs=["output", "mask_index"],
                domain="com.microsoft",
            ),
        ]

        inputs = [
            helper.make_tensor_value_info("input_ids", TensorProto.FLOAT, ["b", "s"]),
            helper.make_tensor_value_info("segment_ids", TensorProto.FLOAT, ["b", "s"]),
        ]

        outputs = [
            helper.make_tensor_value_info("output", TensorProto.FLOAT, None),
            helper.make_tensor_value_info("mask_index", TensorProto.INT32, None),
        ]

        graph = helper.make_graph(nodes, "Unsqueeze_Test", inputs, outputs, initializers)
        model = helper.make_model(graph)

        inferred = SymbolicShapeInference.infer_shapes(model, auto_merge=True)
        expected_shapes = [
            helper.make_tensor_value_info("output", TensorProto.FLOAT, ["b", "s", hidden_size]),
            helper.make_tensor_value_info("mask_index", TensorProto.INT32, ["b"]),
        ]
        self._check_shapes(graph, inferred.graph, expected_shapes)

    def test_softmax_cross_entropy_loss(self):
        hidden_size = 1024

        nodes = [
            helper.make_node("SoftmaxCrossEntropyLoss", inputs=["logits", "labels"], outputs=["loss"]),
        ]

        inputs = [
            helper.make_tensor_value_info("logits", TensorProto.FLOAT, ["b", "s", hidden_size]),
            helper.make_tensor_value_info("labels", TensorProto.INT32, ["b", "s"]),
        ]

        outputs = [
            helper.make_tensor_value_info("loss", TensorProto.FLOAT, None),
        ]

        graph = helper.make_graph(nodes, "SoftmaxCrossEntropyLoss_Test", inputs, outputs, [])
        model = helper.make_model(graph)

        inferred = SymbolicShapeInference.infer_shapes(model, auto_merge=True)
        expected_shapes = [helper.make_tensor_value_info("loss", TensorProto.FLOAT, [])]
        self._check_shapes(graph, inferred.graph, expected_shapes)

    def _test_einsum_one_input_impl(self, input_0_shape, output_0_shape, eqn):
        nodes = [
            helper.make_node("Einsum", ["input_0"], ["output_0"], "einsum_0", equation=eqn),
        ]
        inputs = [
            helper.make_tensor_value_info("input_0", TensorProto.FLOAT, input_0_shape),
        ]
        outputs = [
            helper.make_tensor_value_info("output_0", TensorProto.FLOAT, None),
        ]
        graph = helper.make_graph(nodes, "Einsum_Test", inputs, outputs, [])
        model = helper.make_model(graph)

        inferred = SymbolicShapeInference.infer_shapes(model, auto_merge=True)
        expected_shapes = [helper.make_tensor_value_info("output_0", TensorProto.FLOAT, output_0_shape)]
        self._check_shapes(graph, inferred.graph, expected_shapes)

    def _test_einsum_two_inputs_impl(self, input_0_shape, input_1_shape, output_0_shape, eqn):
        nodes = [
            helper.make_node("Einsum", ["input_0", "input_1"], ["output_0"], "einsum_0", equation=eqn),
        ]
        inputs = [
            helper.make_tensor_value_info("input_0", TensorProto.FLOAT, input_0_shape),
            helper.make_tensor_value_info("input_1", TensorProto.FLOAT, input_1_shape),
        ]
        outputs = [
            helper.make_tensor_value_info("output_0", TensorProto.FLOAT, None),
        ]
        graph = helper.make_graph(nodes, "Einsum_Test", inputs, outputs, [])
        model = helper.make_model(graph)

        inferred = SymbolicShapeInference.infer_shapes(model, auto_merge=True)
        expected_shapes = [helper.make_tensor_value_info("output_0", TensorProto.FLOAT, output_0_shape)]
        self._check_shapes(graph, inferred.graph, expected_shapes)

    def test_einsum_matmul(self):
        self._test_einsum_two_inputs_impl([1, "b", 8], [2, 12, "n"], [1, "b", 12, "n"], "abc, cde -> abde")

    def test_einsum_batch_matmul(self):
        self._test_einsum_two_inputs_impl([5, 2, 3], [5, 3, 4], [5, 2, 4], "bij, bjk -> bik")

    def test_einsum_inner_prod(self):
        self._test_einsum_two_inputs_impl([5], [5], [], "i, i")

    def test_einsum_batch_diagonal(self):
        self._test_einsum_one_input_impl([3, 5, 5], [3, 5], "...ii ->...i")

    def test_einsum_sum(self):
        self._test_einsum_one_input_impl(["a", "b"], ["a"], "ij -> i")

    def test_einsum_transpose(self):
        self._test_einsum_one_input_impl(["a", "b"], ["b", "a"], "ij -> ji")

    def test_mul_precision(self):
        graph_input = onnx.helper.make_tensor_value_info("input", TensorProto.FLOAT, [1024])
        graph_output = onnx.helper.make_tensor_value_info("output", TensorProto.FLOAT, None)

        # initializers
        value = numpy.array([0.5], dtype=numpy.float32)
        constant = numpy_helper.from_array(value, name="constant")

        nodes = [
            # Get the shape of the input tensor: `input_tensor_shape = [1024]`.
            onnx.helper.make_node("Shape", ["input"], ["input_shape"]),
            # mul(1024, 0.5) => 512
            onnx.helper.make_node("Mul", ["input_shape", "constant"], ["output_shape"]),
            # Resize input
            onnx.helper.make_node(
                "Resize", inputs=["input", "", "", "output_shape"], outputs=["output"], mode="nearest"
            ),
        ]

        graph_def = onnx.helper.make_graph(nodes, "TestMulPrecision", [graph_input], [graph_output], [constant])
        model = SymbolicShapeInference.infer_shapes(onnx.helper.make_model(graph_def))
        output_dims = unique_element(model.graph.output).type.tensor_type.shape.dim
        self.assertEqual(len(output_dims), 1)
        self.assertEqual(output_dims[0].dim_value, 512)

    def test_div_precision(self):
        graph_input = onnx.helper.make_tensor_value_info("input", TensorProto.FLOAT, [768])
        graph_output = onnx.helper.make_tensor_value_info("output", TensorProto.FLOAT, None)

        # initializers
        value = numpy.array([1.5], dtype=numpy.float32)
        constant = numpy_helper.from_array(value, name="constant")

        nodes = [
            # Get the shape of the input tensor: `input_tensor_shape = [768]`.
            onnx.helper.make_node("Shape", ["input"], ["input_shape"]),
            # div(768, 1.5) => 512
            onnx.helper.make_node("Div", ["input_shape", "constant"], ["output_shape"]),
            # Resize input
            onnx.helper.make_node(
                "Resize", inputs=["input", "", "", "output_shape"], outputs=["output"], mode="nearest"
            ),
        ]

        graph_def = onnx.helper.make_graph(nodes, "TestDivPrecision", [graph_input], [graph_output], [constant])
        model = SymbolicShapeInference.infer_shapes(onnx.helper.make_model(graph_def))
        output_dims = unique_element(model.graph.output).type.tensor_type.shape.dim
        self.assertEqual(len(output_dims), 1)
        self.assertEqual(output_dims[0].dim_value, 512)

    def test_quantize_linear(self):
        """
        Test ONNX QuantizeLinear op.
        Check that the output shape is propagated from the first input and that the output data
        type comes from the zero-point input.
        """
        initializers = [
            helper.make_tensor(
                "scale",
                TensorProto.FLOAT,
                [],
                [1.0],
            ),
            helper.make_tensor(
                "zero_point",
                TensorProto.INT8,
                [],
                [16],
            ),
        ]

        nodes = [
            helper.make_node(
                "QuantizeLinear",
                inputs=[
                    "input_f32",
                    "scale",
                    "zero_point",
                ],
                outputs=["output_s8"],
            ),
        ]

        inputs = [
            helper.make_tensor_value_info("input_f32", TensorProto.FLOAT, ["b", 2, 3, 4]),
        ]

        outputs = [
            helper.make_tensor_value_info("output_s8", TensorProto.UNDEFINED, None),
        ]

        graph = helper.make_graph(nodes, "QuantizeLinear_Test", inputs, outputs, initializers)
        model = helper.make_model(graph)

        inferred = SymbolicShapeInference.infer_shapes(model, auto_merge=True)

        expected_shapes = [
            helper.make_tensor_value_info("output_s8", TensorProto.INT8, ["b", 2, 3, 4]),
        ]
        self._check_shapes(graph, inferred.graph, expected_shapes)

    def test_quantize_linear_ms_domain(self):
        """
        Test QuantizeLinear op ('com.microsoft' domain).
        Check that the output shape is propagated from the first input and that the output data
        type comes from the zero-point input.
        """
        initializers = [
            helper.make_tensor(
                "scale",
                TensorProto.FLOAT,
                [],
                [1.0],
            ),
            helper.make_tensor(
                "zero_point",
                TensorProto.UINT16,
                [],
                [16],
            ),
        ]

        nodes = [
            helper.make_node(
                "QuantizeLinear",
                inputs=[
                    "input_f32",
                    "scale",
                    "zero_point",
                ],
                outputs=["output_u16"],
                domain="com.microsoft",
            ),
        ]

        inputs = [
            helper.make_tensor_value_info("input_f32", TensorProto.FLOAT, ["b", 2, 3, 4]),
        ]

        outputs = [
            helper.make_tensor_value_info("output_u16", TensorProto.UNDEFINED, None),
        ]

        graph = helper.make_graph(nodes, "QuantizeLinear_MSDomain_Test", inputs, outputs, initializers)
        model = helper.make_model(graph)

        inferred = SymbolicShapeInference.infer_shapes(model, auto_merge=True)

        expected_shapes = [
            helper.make_tensor_value_info("output_u16", TensorProto.UINT16, ["b", 2, 3, 4]),
        ]
        self._check_shapes(graph, inferred.graph, expected_shapes)

    def test_quantize_linear_no_zp_input(self):
        """
        Test QuantizeLinear op ('com.microsoft' domain).
        Check that the output shape is propagated from the first input.
        The zero-point input is missing, so the output data type should default to uint8.
        """
        initializers = [
            helper.make_tensor(
                "scale",
                TensorProto.FLOAT,
                [],
                [1.0],
            ),
        ]

        nodes = [
            helper.make_node(
                "QuantizeLinear",
                inputs=[
                    "input_f32",
                    "scale",
                ],
                outputs=["output_u8"],
                domain="com.microsoft",
            ),
        ]

        inputs = [
            helper.make_tensor_value_info("input_f32", TensorProto.FLOAT, ["b", 2, 3, 4]),
        ]

        outputs = [
            helper.make_tensor_value_info("output_u8", TensorProto.UNDEFINED, None),
        ]

        graph = helper.make_graph(nodes, "QuantizeLinear_NoZP_Test", inputs, outputs, initializers)
        model = helper.make_model(graph)

        inferred = SymbolicShapeInference.infer_shapes(model, auto_merge=True)

        # Check that the output shape is propagated from the first input and that the
        # output data type comes from the zero-point input.
        expected_shapes = [
            helper.make_tensor_value_info("output_u8", TensorProto.UINT8, ["b", 2, 3, 4]),
        ]
        self._check_shapes(graph, inferred.graph, expected_shapes)

    def test_dequantize_linear_ms_domain(self):
        """
        Test DequantizeLinear operator ('com.microsoft' domain).
        Check that the output shape is propagated from the first input and that the output data
        type comes from the scale input.
        """
        initializers = [
            helper.make_tensor(
                "scale",
                TensorProto.FLOAT,
                [],
                [1.0],
            ),
            helper.make_tensor(
                "zero_point",
                TensorProto.UINT16,
                [],
                [16],
            ),
        ]

        nodes = [
            helper.make_node(
                "DequantizeLinear",
                inputs=[
                    "input_u16",
                    "scale",
                    "zero_point",
                ],
                outputs=["output_f32"],
                domain="com.microsoft",
            ),
        ]

        inputs = [
            helper.make_tensor_value_info("input_u16", TensorProto.UINT16, ["b", 2, 3, 4]),
        ]

        outputs = [
            helper.make_tensor_value_info("output_f32", TensorProto.UNDEFINED, None),
        ]

        graph = helper.make_graph(nodes, "DequantizeLinear_MSDomain_Test", inputs, outputs, initializers)
        model = helper.make_model(graph)

        inferred = SymbolicShapeInference.infer_shapes(model, auto_merge=True)

        expected_shapes = [
            helper.make_tensor_value_info("output_f32", TensorProto.FLOAT, ["b", 2, 3, 4]),
        ]
        self._check_shapes(graph, inferred.graph, expected_shapes)

    def test_matmulnbits(self):
        """
        Test ORT MatMulNBits op.
        Check that the output shape is propagated from the inputs and that the output data
        type comes from the first input.
        """
        b_np = numpy.random.randint(0, 255, (4, 1, 8), numpy.uint8)
        b = numpy_helper.from_array(b_np, name="b")
        scale_np = numpy.random.rand(4).astype(numpy.float32)
        scale = numpy_helper.from_array(scale_np, name="scale")
        zero_point_np = numpy.random.randint(0, 255, (4), numpy.uint8)
        zero_point = numpy_helper.from_array(zero_point_np, name="zero_point")

        initializers = [b, scale, zero_point]

        kwargs = {"K": 10, "N": 4, "block_size": 16}

        nodes = [
            helper.make_node(
                "MatMulNBits",
                inputs=[
                    "input_f32",
                    "b",
                    "scale",
                    "zero_point",
                ],
                outputs=["output_f32"],
                **kwargs,
            ),
        ]

        inputs = [
            helper.make_tensor_value_info("input_f32", TensorProto.FLOAT, ["x", 2, 3, 10]),
        ]

        outputs = [
            helper.make_tensor_value_info("output_f32", TensorProto.UNDEFINED, None),
        ]

        graph = helper.make_graph(nodes, "MatMulNBits_Test", inputs, outputs, initializers)
        model = helper.make_model(graph)

        inferred = SymbolicShapeInference.infer_shapes(model, auto_merge=True)

        expected_shapes = [
            helper.make_tensor_value_info("output_f32", TensorProto.FLOAT, ["x", 2, 3, 4]),
        ]
        self._check_shapes(graph, inferred.graph, expected_shapes)


class TestSymbolicShapeInferenceForSlice(unittest.TestCase):
    def check_slice_of_concat(self, input_dims, start, end, step, expected_output_dim):
        _dimstrmap = {dim: f"dim{i}" for i, dim in enumerate(input_dims)}

        def dimstrmap(dim):
            return _dimstrmap.get(dim, dim)

        def get_initializer(name):
            valuemap = {"zero": 0, "one": 1, "two": 2, "ten": 10, "intmax": 2**32}
            value = -valuemap[name[4:]] if name.startswith("neg_") else valuemap[name]
            return onnx.helper.make_tensor(name, TensorProto.INT64, [1], [value])

        initializers = [
            get_initializer(name)
            for name in [
                "zero",
                "one",
                "two",
                "ten",
                "intmax",
                "neg_intmax",
                "neg_one",
                "neg_ten",
            ]
        ]
        inputs = []
        nodes = []
        for i, dim in enumerate(input_dims):
            inputs.append(onnx.helper.make_tensor_value_info(f"t{i}", TensorProto.FLOAT, ["B", dim]))
            nodes.extend(
                [
                    onnx.helper.make_node("Shape", [f"t{i}"], [f"shape{i}"]),
                    onnx.helper.make_node("Slice", [f"shape{i}", "one", "two", "zero", "one"], [f"dim{i}"]),
                    onnx.helper.make_node("Neg", [f"dim{i}"], [f"neg_dim{i}"]),
                ]
            )

        def make_concat_dims(concat_name, dims):
            dims = [f"neg_{dimstrmap(dim[1:])}" if dim.startswith("-") else dimstrmap(dim) for dim in dims]
            return onnx.helper.make_node("Concat", dims, [concat_name], axis=0)

        nodes.extend(
            [
                onnx.helper.make_node("Concat", [inp.name for inp in inputs], ["concat"], axis=1),
                make_concat_dims("starts", ["zero", start]),
                make_concat_dims("ends", ["intmax", end]),
                make_concat_dims("axes", ["zero", "one"]),
                make_concat_dims("steps", ["one", step]),
                onnx.helper.make_node("Slice", ["concat", "starts", "ends", "axes", "steps"], ["output"]),
            ]
        )
        output = onnx.helper.make_tensor_value_info("output", TensorProto.FLOAT, ["d1", "d2"])
        graph_def = onnx.helper.make_graph(nodes, "graph", inputs, [output], initializer=initializers)
        model = SymbolicShapeInference.infer_shapes(onnx.helper.make_model(graph_def))
        output = unique_element(model.graph.output)
        shape = [d.dim_param if d.dim_param else d.dim_value for d in output.type.tensor_type.shape.dim]
        self.assertEqual(shape, ["B", expected_output_dim])

    def test_numeric_negative_indices_forward(self):
        self.check_slice_of_concat(["M"], "-ten", "-one", "one", 9)

    def test_numeric_negative_indices_backward(self):
        self.check_slice_of_concat(["M"], "-one", "-ten", "-one", 9)

    def test_symbolic_end_index(self):
        self.check_slice_of_concat(["M", "N"], "zero", "M", "one", "M")

    def test_symbolic_negative_start_index(self):
        self.check_slice_of_concat(["M", "N"], "-N", "intmax", "one", "N")

    def test_non_unit_step(self):
        self.check_slice_of_concat(["N", "N"], "zero", "intmax", "two", "N")

    def test_symbolic_step(self):
        self.check_slice_of_concat(["N", "N"], "zero", "intmax", "N", "floor(-1/N) + 3")

    def test_symbolic_negative_step(self):
        self.check_slice_of_concat(["N", "N"], "-one", "-intmax", "-N", "floor(-1/N) + 3")

    def test_flip(self):
        self.check_slice_of_concat(["N"], "-one", "-intmax", "-one", "N")

    def test_flip_of_concat(self):
        self.check_slice_of_concat(["N", "N", "N"], "-one", "-intmax", "-one", "3*N")

    def test_slice_of_min(self):
        graph_input = onnx.helper.make_tensor_value_info("input", TensorProto.FLOAT, ["N"])
        graph_output = onnx.helper.make_tensor_value_info("output", TensorProto.FLOAT, None)
        const_size = 42
        half_const_size = const_size // 2
        initializers = [
            onnx.helper.make_tensor("const_tensor", TensorProto.FLOAT, [const_size], [42.0] * const_size),
            onnx.helper.make_tensor("half_const_size", TensorProto.INT64, [], [half_const_size]),
            onnx.helper.make_tensor("zeros", TensorProto.INT64, [1], [0]),
            onnx.helper.make_tensor("ones", TensorProto.INT64, [1], [1]),
        ]

        nodes = [
            # Get the shape of the input tensor: `input_tensor_shape = [N]`.
            onnx.helper.make_node("Shape", ["input"], ["input_tensor_shape"]),
            # The starts of the const tensor slice: `starts = [21 - N + 1]`.
            onnx.helper.make_node("Sub", ["half_const_size", "input_tensor_shape"], ["starts_aux"]),
            onnx.helper.make_node("Add", ["ones", "starts_aux"], ["starts"]),
            # The ends of the const tensor slice: `ends = [21 + N]`.
            onnx.helper.make_node("Add", ["half_const_size", "input_tensor_shape"], ["ends"]),
            # Slice the const tensor: `slice_out = const_tensor[starts:ends]`.
            onnx.helper.make_node("Slice", ["const_tensor", "starts", "ends", "zeros", "ones"], ["slice_out"]),
            # Crop the const tensor slice using the shape of the input tensor: `slice_out_cropped = slice_out[0:input_tensor_shape]`.
            onnx.helper.make_node(
                "Slice", ["slice_out", "zeros", "input_tensor_shape", "zeros", "ones"], ["slice_out_cropped"]
            ),
            # Add the const tensor slice to the input tensor: `output = input + slice_out_cropped`.
            onnx.helper.make_node("Add", ["slice_out_cropped", "input"], ["output"]),
        ]

        graph_def = onnx.helper.make_graph(nodes, "SliceOfMin", [graph_input], [graph_output], initializer=initializers)
        model = SymbolicShapeInference.infer_shapes(onnx.helper.make_model(graph_def))
        output_dims = unique_element(model.graph.output).type.tensor_type.shape.dim
        self.assertEqual(len(output_dims), 1)
        self.assertEqual(output_dims[0].dim_param, "N")


if __name__ == "__main__":
    unittest.main()



import onnx
from onnx import TensorProto, helper


# Generate a basic QDQ Conv model with `num_convs` Conv nodes and their surrounding DQ/Q nodes
def GenerateModel(model_path, num_convs):  # noqa: N802
    nodes = []
    initializers = []
    inputs = []
    outputs = []

    for i in range(num_convs):

        def name(base):
            return f"{base}_{i}"  # noqa: B023

        nodes.extend(
            [
                helper.make_node(
                    "DequantizeLinear",
                    [name("X"), name("Scale"), name("Zero_point_uint8")],
                    [name("input_DQ")],
                    name("input_DQ"),
                ),
                helper.make_node(
                    "DequantizeLinear",
                    [name("W"), name("Scale"), name("Zero_point_uint8")],
                    [name("conv_weight_DQ")],
                    name("conv_weight_DQ"),
                ),
                helper.make_node(
                    "DequantizeLinear",
                    [name("Bias"), name("Scale"), name("Zero_point_int32")],
                    [name("conv_bias_DQ")],
                    name("conv_bias_DQ"),
                ),
                helper.make_node(
                    "Conv",
                    [name("input_DQ"), name("conv_weight_DQ"), name("conv_bias_DQ")],
                    [name("conv_output")],
                    name("conv"),
                ),
                helper.make_node(
                    "QuantizeLinear",
                    [name("conv_output"), name("Scale"), name("Zero_point_uint8")],
                    [name("Y")],
                    name("output_Q"),
                ),
            ]
        )

        initializers.extend(
            [
                helper.make_tensor(name("Scale"), TensorProto.FLOAT, [1], [256.0]),
                helper.make_tensor(name("Zero_point_uint8"), TensorProto.UINT8, [1], [0]),
                helper.make_tensor(name("Zero_point_int32"), TensorProto.INT32, [1], [0]),
                helper.make_tensor(name("W"), TensorProto.UINT8, [1, 1, 3, 3], [128] * 9),
                helper.make_tensor(name("Bias"), TensorProto.INT32, [1], [64]),
            ]
        )

        inputs.extend(
            [
                helper.make_tensor_value_info(name("X"), TensorProto.UINT8, [1, 1, 5, 5]),
            ]
        )

        outputs.extend(
            [
                helper.make_tensor_value_info(name("Y"), TensorProto.UINT8, [1, 1, 3, 3]),
            ]
        )

    graph = helper.make_graph(nodes, f"QDQ_Conv_x_{num_convs}", inputs, outputs, initializers)

    model = helper.make_model(graph)
    onnx.save(model, model_path)


if __name__ == "__main__":
    GenerateModel("qdq_conv.onnx", 1)
    GenerateModel("runtime_optimization/qdq_convs.onnx", 3)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import unittest

import numpy as np
from onnx import TensorProto, helper, numpy_helper, save
from op_test_utils import (
    check_model_correctness,
    check_op_type_count,
    check_qtype_by_node_type,
    input_feeds_neg_one_zero_one,
)

from onnxruntime.quantization import QuantFormat, QuantType, quantize_static


class TestConcatModel(unittest.TestCase):
    def construct_model(self, model_path):
        #          (input)
        #         /    |  \
        #        /     |   \
        #       /      |    \
        #      /       |     \
        #  Conv(1)  Conv(2)  conv(3)
        #       \      |     /
        #         \    |    /
        #           \  |   /
        #            Concat
        #              |
        #             Relu
        #              |
        #           Identity
        #              |
        #           (output)
        initializers = []
        input = helper.make_tensor_value_info("input", TensorProto.FLOAT, [1, 3, 15, 15])
        output = helper.make_tensor_value_info("output", TensorProto.FLOAT, [1, 13, 13, 13])

        # Conv1 output [1, 2, 13, 13]
        conv1_weight_initializer = numpy_helper.from_array(
            np.random.randint(-1, 2, [2, 3, 3, 3]).astype(np.float32),
            name="conv1_weight",
        )
        conv1_node = helper.make_node("Conv", ["input", "conv1_weight"], ["conv1_output"], name="conv1_node")

        # Conv2 output [1, 5, 13, 13]
        conv2_weight_initializer = numpy_helper.from_array(
            np.random.randint(-1, 2, [5, 3, 3, 3]).astype(np.float32),
            name="conv2_weight",
        )
        conv2_node = helper.make_node("Conv", ["input", "conv2_weight"], ["conv2_output"], name="conv2_node")

        # Conv3 output [1, 6, 13, 13]
        conv3_weight_initializer = numpy_helper.from_array(
            np.random.randint(-1, 2, [6, 3, 3, 3]).astype(np.float32),
            name="conv3_weight",
        )
        conv3_node = helper.make_node("Conv", ["input", "conv3_weight"], ["conv3_output"], name="conv3_node")

        concat_node = helper.make_node(
            "Concat",
            ["conv1_output", "conv2_output", "conv3_output"],
            ["concat_output"],
            name="concat_node",
            axis=1,
        )

        relu_node = helper.make_node("Relu", ["concat_output"], ["relu_output"], name="relu_node")
        identity_node = helper.make_node("Identity", ["relu_output"], ["output"], name="identity_node")

        initializers = [
            conv1_weight_initializer,
            conv2_weight_initializer,
            conv3_weight_initializer,
        ]
        graph = helper.make_graph(
            [conv1_node, conv2_node, conv3_node, concat_node, relu_node, identity_node],
            "qlinear_concat_op_test",
            [input],
            [output],
            initializer=initializers,
        )
        model = helper.make_model(graph, opset_imports=[helper.make_opsetid("", 13)])
        save(model, model_path)

    def quantize_concat_test(self, activation_type, weight_type, extra_options={}):  # noqa: B006
        np.random.seed(1)
        model_fp32_path = "concat_fp32.onnx"
        self.construct_model(model_fp32_path)
        data_reader = input_feeds_neg_one_zero_one(1, {"input": [1, 3, 15, 15]})

        activation_proto_qtype = TensorProto.UINT8 if activation_type == QuantType.QUInt8 else TensorProto.INT8
        activation_type_str = "u8" if (activation_type == QuantType.QUInt8) else "s8"
        weight_type_str = "u8" if (weight_type == QuantType.QUInt8) else "s8"
        model_q8_path = f"concat_{activation_type_str}{weight_type_str}.onnx"
        model_q8_qdq_path = f"concat_{activation_type_str}{weight_type_str}_qdq.onnx"

        # Verify QOperator mode
        data_reader.rewind()
        quantize_static(
            model_fp32_path,
            model_q8_path,
            data_reader,
            quant_format=QuantFormat.QOperator,
            activation_type=activation_type,
            weight_type=weight_type,
            extra_options=extra_options,
        )

        qnode_counts = {
            "QLinearConv": 3,
            "QuantizeLinear": 1,
            "DequantizeLinear": 1,
            "QLinearConcat": 1,
        }
        check_op_type_count(self, model_q8_path, **qnode_counts)
        qnode_io_qtypes = {
            "QuantizeLinear": [
                ["i", 2, activation_proto_qtype],
                ["o", 0, activation_proto_qtype],
            ]
        }
        qnode_io_qtypes.update(
            {
                "QLinearConcat": [
                    ["i", 1, activation_proto_qtype],
                    ["i", 4, activation_proto_qtype],
                    ["i", 7, activation_proto_qtype],
                ]
            }
        )
        check_qtype_by_node_type(self, model_q8_path, qnode_io_qtypes)
        data_reader.rewind()
        check_model_correctness(self, model_fp32_path, model_q8_path, data_reader.get_next())

        # Verify QDQ mode
        data_reader.rewind()
        quantize_static(
            model_fp32_path,
            model_q8_qdq_path,
            data_reader,
            quant_format=QuantFormat.QDQ,
            activation_type=activation_type,
            weight_type=weight_type,
            extra_options=extra_options,
        )

        relu_count = 0
        if activation_type == QuantType.QInt8 and extra_options.get("ActivationSymmetric", False):
            relu_count = 1
        qdqnode_counts = {
            "Conv": 3,
            "QuantizeLinear": 5 + relu_count,
            "DequantizeLinear": 8 + relu_count,
            "Concat": 1,
            "Relu": 0 + relu_count,
        }
        check_op_type_count(self, model_q8_qdq_path, **qdqnode_counts)
        qnode_io_qtypes = {
            "QuantizeLinear": [
                ["i", 2, activation_proto_qtype],
                ["o", 0, activation_proto_qtype],
            ]
        }
        check_qtype_by_node_type(self, model_q8_qdq_path, qnode_io_qtypes)
        data_reader.rewind()
        check_model_correctness(self, model_fp32_path, model_q8_qdq_path, data_reader.get_next())

    def test_quantize_concat(self):
        self.quantize_concat_test(QuantType.QUInt8, QuantType.QUInt8, extra_options={})

    def test_quantize_concat_s8s8(self):
        self.quantize_concat_test(
            QuantType.QInt8,
            QuantType.QInt8,
            extra_options={"ActivationSymmetric": True},
        )


if __name__ == "__main__":
    unittest.main()



from pathlib import Path

import onnx
from onnx import helper
from onnx.onnx_pb import TensorProto

nodes = [
    helper.make_node("Gather", ["X", "indices"], ["Y"], axis=1),
    helper.make_node("Shape", ["Y"], ["Y_shape"]),
]

graph = helper.make_graph(
    nodes,
    "GatherWithScalarIndicesThenShape",
    [  # input
        helper.make_tensor_value_info("X", TensorProto.FLOAT, ["M", "N", "K"]),
        helper.make_tensor_value_info("indices", TensorProto.INT64, []),
    ],
    [  # output
        helper.make_tensor_value_info("Y", TensorProto.FLOAT, None),
        helper.make_tensor_value_info("Y_shape", TensorProto.INT64, None),
    ],
)

model = helper.make_model(graph, opset_imports=[helper.make_opsetid("", 19)])
onnx.save(model, str(Path(__file__).with_suffix(".onnx")))



#!/usr/bin/env python3
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
from __future__ import annotations

import argparse
import logging
import os
import subprocess
import warnings

log = logging.getLogger("Build")


def parse_nightly_and_local_version_from_whl_name(blob_name):
    night_build = "nightly" if blob_name.find(".dev") > 0 else "stable"

    start = blob_name.find("+")
    if start == -1:
        return night_build, None
    start = start + 1
    end = blob_name.find("-", start)
    if end == -1:
        return night_build, None
    return night_build, blob_name[start:end]


def run_subprocess(args, cwd=None):
    log.warning(f"Running subprocess in '{cwd or os.getcwd()}'\n{args}")
    return subprocess.run(args, cwd=cwd, check=True)


def upload_whl(python_wheel_path, final_storage=False):
    storage_account_name = "onnxruntimepackages" if final_storage else "onnxruntimepackagesint"
    blob_name = os.path.basename(python_wheel_path)
    run_subprocess(["azcopy", "cp", python_wheel_path, f"https://{storage_account_name}.blob.core.windows.net/$web/"])

    nightly_build, local_version = parse_nightly_and_local_version_from_whl_name(blob_name)
    if local_version:
        html_blob_name = f"onnxruntime_{nightly_build}_{local_version}.html"
    else:
        html_blob_name = f"onnxruntime_{nightly_build}.html"

    download_path_to_html = f"./onnxruntime_{nightly_build}.html"

    run_subprocess(
        [
            "azcopy",
            "cp",
            f"https://{storage_account_name}.blob.core.windows.net/$web/" + html_blob_name,
            download_path_to_html,
        ]
    )

    blob_name_plus_replaced = blob_name.replace("+", "%2B")
    with open(download_path_to_html) as f:
        lines = f.read().splitlines()

    new_line = f'<a href="{blob_name_plus_replaced}">{blob_name_plus_replaced}</a><br>'
    if new_line not in lines:
        lines.append(new_line)
        lines.sort()

        with open(download_path_to_html, "w") as f:
            for item in lines:
                f.write(f"{item}\n")
    else:
        warnings.warn(f"'{new_line}' exists in {download_path_to_html}. The html file is not updated.")
    run_subprocess(
        [
            "azcopy",
            "cp",
            download_path_to_html,
            f"https://{storage_account_name}.blob.core.windows.net/$web/" + html_blob_name,
            "--content-type",
            "text/html",
            "--overwrite",
            "true",
        ]
    )


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Upload python whl to azure storage.")

    parser.add_argument("--python_wheel_path", type=str, help="path to python wheel")
    parser.add_argument("--final_storage", action="store_true", help="upload to final storage")

    args = parser.parse_args()

    upload_whl(args.python_wheel_path, args.final_storage)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# __init__.py

from .sampler import LoadBalancingDistributedBatchSampler, LoadBalancingDistributedSampler  # noqa: F401



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation.  All rights reserved.
# Licensed under the MIT License.  See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import os
import unittest

import numpy as np
import onnx
from gpt2_model_generator import create_gpt2_embedlayer
from parity_utilities import find_transformers_source

from onnxruntime import InferenceSession

if find_transformers_source():
    from onnx_model import OnnxModel
    from optimizer import optimize_model
else:
    from onnxruntime.transformers.onnx_model import OnnxModel
    from onnxruntime.transformers.optimizer import optimize_model


class TestFusion(unittest.TestCase):
    def verify_fusion(self, optimized_model, expected_model_filename):
        optimized_model.topological_sort()

        expected_model_path = os.path.join(os.path.dirname(__file__), "test_data", "models", expected_model_filename)
        expected_model = OnnxModel(onnx.load(expected_model_path))
        expected_model.topological_sort()

        self.assertEqual(str(optimized_model.model.graph), str(expected_model.model.graph))

    def verify_parity(self, optimized_model_path, expected_model_filename):
        expected_model_path = os.path.join(os.path.dirname(__file__), "test_data", "models", expected_model_filename)
        sess_optimized = InferenceSession(optimized_model_path, providers=["CPUExecutionProvider"])
        sess_expected = InferenceSession(expected_model_path, providers=["CPUExecutionProvider"])
        inputs = np.random.randint(low=0, high=6, size=(4, 8), dtype=np.int32) + 1

        outputs_optimized = sess_optimized.run(None, {"ids": inputs})
        outputs_expected = sess_expected.run(None, {"ids": inputs})
        self.assertTrue(np.allclose(outputs_optimized[0], outputs_expected[0]))

    def test_embedlayer_fusion(self):
        model = create_gpt2_embedlayer(one_attention_node=False)
        path = "."
        original_model_path = os.path.join(path, "gpt2_embedlayer.onnx")
        optimized_model_path = os.path.join(path, "gpt2_embedlayer_opt.onnx")
        expected_model_filename = "gpt2_embedlayer_exp.onnx"

        onnx.save(model, original_model_path)
        optimized_model = optimize_model(original_model_path, model_type="gpt2")
        optimized_model.save_model_to_file(optimized_model_path, use_external_data_format=True)

        self.verify_fusion(optimized_model, expected_model_filename)
        self.verify_parity(optimized_model_path, expected_model_filename)
        os.remove(original_model_path)
        os.remove(optimized_model_path)

    def test_embedlayer_fusion_one_attn_node(self):
        model = create_gpt2_embedlayer(one_attention_node=True)
        path = "."
        original_model_path = os.path.join(path, "gpt2_embedlayer_one_attn.onnx")
        optimized_model_path = os.path.join(path, "gpt2_embedlayer_one_attn_opt.onnx")
        expected_model_filename = "gpt2_embedlayer_one_attn_exp.onnx"

        onnx.save(model, original_model_path)
        optimized_model = optimize_model(original_model_path, model_type="gpt2")
        optimized_model.save_model_to_file(optimized_model_path, use_external_data_format=True)

        self.verify_fusion(optimized_model, expected_model_filename)
        self.verify_parity(optimized_model_path, expected_model_filename)
        os.remove(original_model_path)
        os.remove(optimized_model_path)

    def test_embedlayer_fusion_with_embedding_sum_output(self):
        model = create_gpt2_embedlayer(one_attention_node=True, output_embedding_sum=True)
        path = "."
        original_model_path = os.path.join(path, "gpt2_embedlayer_one_attn_output_sum.onnx")
        optimized_model_path = os.path.join(path, "gpt2_embedlayer_one_attn_output_sum_opt.onnx")
        expected_model_filename = "gpt2_embedlayer_one_attn_output_sum_exp.onnx"

        onnx.save(model, original_model_path)
        optimized_model = optimize_model(original_model_path, model_type="gpt2")
        optimized_model.save_model_to_file(optimized_model_path, use_external_data_format=True)

        self.verify_fusion(optimized_model, expected_model_filename)
        self.verify_parity(optimized_model_path, expected_model_filename)
        os.remove(original_model_path)
        os.remove(optimized_model_path)

    def test_embedlayer_fusion_with_embedding_sum_output_no_sln(self):
        model = create_gpt2_embedlayer(one_attention_node=True, has_skip_layer_norm=False, output_embedding_sum=True)
        path = "."
        original_model_path = os.path.join(path, "gpt2_embedlayer_one_attn_output_sum_no_sln.onnx")
        optimized_model_path = os.path.join(path, "gpt2_embedlayer_one_attn_output_sum_no_sln_opt.onnx")
        expected_model_filename = "gpt2_embedlayer_one_attn_output_sum_exp.onnx"

        onnx.save(model, original_model_path)
        optimized_model = optimize_model(original_model_path, model_type="gpt2")
        optimized_model.save_model_to_file(optimized_model_path, use_external_data_format=True)

        self.verify_fusion(optimized_model, expected_model_filename)
        self.verify_parity(optimized_model_path, expected_model_filename)
        os.remove(original_model_path)
        os.remove(optimized_model_path)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import argparse
import pathlib
import shutil
import subprocess
import tempfile

SCRIPT_DIR = pathlib.Path(__file__).parent.resolve()


def update_schema_names(schema_path: pathlib.Path, updated_schema_path: pathlib.Path, names_to_update: dict[str, str]):
    """Replace the occurrence of every key in names_to_update with the corresponding value

    This function parses the schema file at schema_path line by line and replaces the occurrence of
    every key in names_to_update with the corresponding value. The updated schema is written to updated_schema_path.

    Args:
        schema_path (pathlib.Path): Input schema file path.
        updated_schema_path (pathlib.Path): Output schema file path.
        names_to_update (dict[str, str]): Dictionary of names to update.
                                          The key is the name to replace and the value is the new name.
    """
    with open(schema_path) as input, open(updated_schema_path, "w") as output:
        for line in input:
            updated_line = line
            for old_name, new_name in names_to_update.items():
                updated_line = updated_line.replace(old_name, new_name)
            output.write(updated_line)


def generate_python(flatc: pathlib.Path, schema_path: pathlib.Path, output_dir: pathlib.Path):
    # run flatc to generate Python code
    cmd = [str(flatc), "--python", str(schema_path)]
    subprocess.run(cmd, check=True, cwd=output_dir)


def create_init_py(output_dir: pathlib.Path):
    # create an __init__.py that imports all the py files so we can just 'import ort_flatbuffers_py.fbs'
    # in a script that wants to process an ORT format model
    init_py_path = output_dir / "ort_flatbuffers_py/fbs/__init__.py"
    with open(init_py_path, "w") as init_py:
        init_py.write(
            """from os.path import dirname, basename, isfile, join, splitext
import glob
modules = glob.glob(join(dirname(__file__), "*.py"))
__all__ = [splitext(basename(f))[0] for f in modules if isfile(f) and not f.endswith('__init__.py')]

from . import *
"""
        )


def generate_cpp(flatc: pathlib.Path, schema_path: pathlib.Path):
    # run flatc to generate C++ code
    cmd = [str(flatc), "--cpp", "--scoped-enums", "--filename-suffix", ".fbs", str(schema_path)]
    subprocess.run(cmd, check=True, cwd=SCRIPT_DIR)


def main():
    parser = argparse.ArgumentParser(
        description="Generate language bindings for the ORT flatbuffers schema.",
        usage="Provide the path to the flatbuffers flatc executable. "
        "Script can be executed from anywhere but must be located in its original "
        "directory in the ONNX Runtime enlistment.",
    )

    parser.add_argument(
        "-f",
        "--flatc",
        required=True,
        type=pathlib.Path,
        help="Path to flatbuffers flatc executable. "
        "Can be found in the build directory under _deps/flatbuffers-build/<config>/",
    )

    all_languages = ["python", "cpp"]
    parser.add_argument(
        "-l",
        "--language",
        action="append",
        dest="languages",
        choices=all_languages,
        help="Specify which language bindings to generate.",
    )

    args = parser.parse_args()
    languages = args.languages if args.languages is not None else all_languages
    flatc = args.flatc.resolve(strict=True)
    schema_path = SCRIPT_DIR / "ort.fbs"
    training_schema_path = SCRIPT_DIR / "ort_training_checkpoint.fbs"

    test_dir = SCRIPT_DIR.parents[2] / "test" / "flatbuffers"
    test_schema = "flatbuffers_utils_test.fbs"
    test_schema_path = test_dir / test_schema

    if "python" in languages:
        with tempfile.TemporaryDirectory() as temp_dir_name:
            temp_dir = pathlib.Path(temp_dir_name).resolve()
            updated_schema_path = temp_dir / "ort.py.fbs"
            # Create a copy of the schema so we can replace the namespace so that the generated module name doesn't
            # clash with the 'onnxruntime' package.
            # We do this for both the ort format as well as the training checkpoint format schema.
            # We convert any line with the namespace to use ort_flatbuffers_py instead of onnxruntime as the top level
            # namespace. This doesn't change how anything works - it just avoids a naming clash with the 'real'
            # onnxruntime python package
            update_schema_names(
                schema_path, updated_schema_path, {"namespace onnxruntime.fbs;": "namespace ort_flatbuffers_py.fbs;"}
            )

            output_dir = temp_dir / "out"
            output_dir.mkdir()
            generate_python(flatc, updated_schema_path, output_dir)

            updated_training_schema_path = temp_dir / "ort_training_checkpoint.py.fbs"
            # Also replace the `include "ort.fbs";` with `include "ort.py.fbs";` since we update the file name
            # that is written to in the previous step.
            update_schema_names(
                training_schema_path,
                updated_training_schema_path,
                {
                    "namespace onnxruntime.fbs;": "namespace ort_flatbuffers_py.fbs;",
                    'include "ort.fbs";': 'include "ort.py.fbs";',
                },
            )
            generate_python(flatc, updated_training_schema_path, output_dir)
            create_init_py(output_dir)

            # replace generated files in repo
            target_dir = SCRIPT_DIR.parent / "ort_flatbuffers_py"
            if target_dir.is_dir():
                shutil.rmtree(target_dir)
            shutil.move(str(output_dir / "ort_flatbuffers_py"), str(target_dir))

    if "cpp" in languages:
        generate_cpp(flatc, schema_path)
        generate_cpp(flatc, training_schema_path)

        # generate schema used in unit tests and move to the test dir. file is generated by generate_cpp in SCRIPT_DIR
        generate_cpp(flatc, test_schema_path)
        src_test_schema_h = str(SCRIPT_DIR / test_schema) + ".h"
        dest_test_schema_h = str(test_dir / test_schema) + ".h"
        shutil.move(src_test_schema_h, dest_test_schema_h)


if __name__ == "__main__":
    main()



import onnx
from onnx import OperatorSetIdProto, TensorProto, helper

onnxdomain = OperatorSetIdProto()
onnxdomain.version = 12
# The empty string ("") or absence of this field implies the operator set that is defined as part of the ONNX specification.
onnxdomain.domain = ""
msdomain = OperatorSetIdProto()
msdomain.version = 1
msdomain.domain = "com.microsoft"
opsets = [onnxdomain, msdomain]


def save(model_path, nodes, inputs, outputs, initializers):
    graph = helper.make_graph(nodes, "GemmSumTest", inputs, outputs, initializers)

    model = helper.make_model(graph, opset_imports=opsets, producer_name="onnxruntime-test")

    print(model_path)
    onnx.save(model, model_path)


def gen_gemm_sum_basic(model_path):
    nodes = [
        helper.make_node(op_type="Gemm", inputs=["A", "B"], outputs=["tp0"]),
        helper.make_node(op_type="Sum", inputs=["tp0", "C"], outputs=["output"]),
    ]

    inputs = [
        helper.make_tensor_value_info("A", TensorProto.FLOAT, ["M", "K"]),
        helper.make_tensor_value_info("B", TensorProto.FLOAT, ["K", "N"]),
        helper.make_tensor_value_info("C", TensorProto.FLOAT, ["M", "N"]),
    ]

    outputs = [helper.make_tensor_value_info("output", TensorProto.FLOAT, ["M", "N"])]

    save(model_path, nodes, inputs, outputs, initializers=[])


def gen_gemm_sum_attributes(model_path):
    nodes = [
        helper.make_node(
            op_type="Gemm",
            inputs=["A", "B"],
            outputs=["tp0"],
            alpha=3.5,
            beta=6.25,
            transA=False,
            transB=True,
        ),
        helper.make_node(op_type="Sum", inputs=["tp0", "C"], outputs=["output"]),
    ]

    inputs = [
        helper.make_tensor_value_info("A", TensorProto.FLOAT, ["M", "K"]),
        helper.make_tensor_value_info("B", TensorProto.FLOAT, ["N", "K"]),
        helper.make_tensor_value_info("C", TensorProto.FLOAT, ["M", "N"]),
    ]

    outputs = [helper.make_tensor_value_info("output", TensorProto.FLOAT, ["M", "N"])]

    save(model_path, nodes, inputs, outputs, initializers=[])


def gen_gemm_sum_internal_nodes(model_path):
    nodes = [
        helper.make_node(op_type="Identity", inputs=["A"], outputs=["tp0"]),
        helper.make_node(op_type="Identity", inputs=["B"], outputs=["tp1"]),
        helper.make_node(op_type="Gemm", inputs=["tp0", "tp1"], outputs=["tp2"]),
        helper.make_node(op_type="Identity", inputs=["C"], outputs=["tp3"]),
        helper.make_node(op_type="Sum", inputs=["tp2", "tp3"], outputs=["tp4"]),
        helper.make_node(op_type="Identity", inputs=["tp4"], outputs=["output"]),
    ]

    inputs = [
        helper.make_tensor_value_info("A", TensorProto.FLOAT, ["M", "K"]),
        helper.make_tensor_value_info("B", TensorProto.FLOAT, ["K", "N"]),
        helper.make_tensor_value_info("C", TensorProto.FLOAT, ["M", "N"]),
    ]

    outputs = [helper.make_tensor_value_info("output", TensorProto.FLOAT, ["M", "N"])]

    save(model_path, nodes, inputs, outputs, initializers=[])


def gen_gemm_sum_no_fusion_c_used(model_path):
    nodes = [
        helper.make_node(op_type="Gemm", inputs=["A", "B", "C"], outputs=["tp0"]),
        helper.make_node(op_type="Sum", inputs=["tp0", "D"], outputs=["output"]),
    ]

    inputs = [
        helper.make_tensor_value_info("A", TensorProto.FLOAT, ["M", "K"]),
        helper.make_tensor_value_info("B", TensorProto.FLOAT, ["K", "N"]),
        helper.make_tensor_value_info("C", TensorProto.FLOAT, ["M", "N"]),
        helper.make_tensor_value_info("D", TensorProto.FLOAT, ["M", "N"]),
    ]

    outputs = [helper.make_tensor_value_info("output", TensorProto.FLOAT, ["M", "N"])]

    save(model_path, nodes, inputs, outputs, initializers=[])


def gen_gemm_sum_no_fusion_sum_multiple_inputs(model_path):
    nodes = [
        helper.make_node(op_type="Gemm", inputs=["A", "B"], outputs=["tp0"]),
        helper.make_node(op_type="Sum", inputs=["tp0", "C", "D"], outputs=["output"]),
    ]

    inputs = [
        helper.make_tensor_value_info("A", TensorProto.FLOAT, ["M", "K"]),
        helper.make_tensor_value_info("B", TensorProto.FLOAT, ["K", "N"]),
        helper.make_tensor_value_info("C", TensorProto.FLOAT, ["M", "N"]),
        helper.make_tensor_value_info("D", TensorProto.FLOAT, ["M", "N"]),
    ]

    outputs = [helper.make_tensor_value_info("output", TensorProto.FLOAT, ["M", "N"])]

    save(model_path, nodes, inputs, outputs, initializers=[])


def gen_gemm_sum_fusion_broadcast(model_path):
    nodes = [
        helper.make_node(op_type="Gemm", inputs=["A", "B"], outputs=["tp0"]),
        helper.make_node(op_type="Sum", inputs=["tp0", "C"], outputs=["output"]),
    ]

    inputs = [
        helper.make_tensor_value_info("A", TensorProto.FLOAT, ["M", "K"]),
        helper.make_tensor_value_info("B", TensorProto.FLOAT, ["K", "N"]),
        helper.make_tensor_value_info("C", TensorProto.FLOAT, [1, "N"]),
    ]

    outputs = [helper.make_tensor_value_info("output", TensorProto.FLOAT, ["M", "N"])]

    save(model_path, nodes, inputs, outputs, initializers=[])


def gen_gemm_sum_no_fusion_broadcast_failure(model_path):
    nodes = [
        helper.make_node(op_type="Gemm", inputs=["A", "B"], outputs=["tp0"]),
        helper.make_node(op_type="Sum", inputs=["tp0", "C"], outputs=["output"]),
    ]

    inputs = [
        helper.make_tensor_value_info("A", TensorProto.FLOAT, ["M", "K"]),
        helper.make_tensor_value_info("B", TensorProto.FLOAT, ["K", "N"]),
        # should work with multidirectional broadcast as second argument, but not unidirectional.
        helper.make_tensor_value_info("C", TensorProto.FLOAT, [1, "M", "N"]),
    ]

    outputs = [helper.make_tensor_value_info("output", TensorProto.FLOAT, [1, "M", "N"])]

    save(model_path, nodes, inputs, outputs, initializers=[])


def gen_gemm_sum_no_fusion_original_gemm_output_used(model_path):
    nodes = [
        helper.make_node(op_type="Gemm", inputs=["A", "B"], outputs=["tp0"]),
        helper.make_node(op_type="Sum", inputs=["tp0", "C"], outputs=["output"]),
    ]

    inputs = [
        helper.make_tensor_value_info("A", TensorProto.FLOAT, ["M", "K"]),
        helper.make_tensor_value_info("B", TensorProto.FLOAT, ["K", "N"]),
        helper.make_tensor_value_info("C", TensorProto.FLOAT, ["M", "N"]),
    ]

    outputs = [
        helper.make_tensor_value_info("tp0", TensorProto.FLOAT, ["M", "N"]),
        helper.make_tensor_value_info("output", TensorProto.FLOAT, ["M", "N"]),
    ]

    save(model_path, nodes, inputs, outputs, initializers=[])


gen_gemm_sum_basic("gemm_sum_basic.onnx")
gen_gemm_sum_attributes("gemm_sum_attributes.onnx")
gen_gemm_sum_internal_nodes("gemm_sum_internal_nodes.onnx")
gen_gemm_sum_no_fusion_c_used("gemm_sum_no_fusion_c_used.onnx")
gen_gemm_sum_no_fusion_sum_multiple_inputs("gemm_sum_no_fusion_sum_multiple_inputs.onnx")
gen_gemm_sum_fusion_broadcast("gemm_sum_fusion_broadcast.onnx")
gen_gemm_sum_no_fusion_broadcast_failure("gemm_sum_no_fusion_broadcast_failure.onnx")
gen_gemm_sum_no_fusion_original_gemm_output_used("gemm_sum_no_fusion_original_gemm_output_used.onnx")



import itertools

import numpy as np
import onnx
from onnx import OperatorSetIdProto, TensorProto, helper

onnxdomain = OperatorSetIdProto()
onnxdomain.version = 12
# The empty string ("") or absence of this field implies the operator set that is defined as part of the ONNX specification.
onnxdomain.domain = ""
msdomain = OperatorSetIdProto()
msdomain.version = 1
msdomain.domain = "com.microsoft"
opsets = [onnxdomain, msdomain]

# expect type to be either TensorProto.FLOAT or TensorProto.FLOAT16


def type_to_string(type):
    return "float" if type == TensorProto.FLOAT else "float16"


def save(model_path, nodes, inputs, outputs, initializers):
    graph = helper.make_graph(nodes, "CastPropagateTest", inputs, outputs, initializers)

    model = helper.make_model(graph, opset_imports=opsets, producer_name="onnxruntime-test")

    onnx.save(model, model_path + ".onnx")


def gen_fuse_back2back_casts(model_path):
    for type1, type2 in list(itertools.product([TensorProto.FLOAT, TensorProto.FLOAT16], repeat=2)):
        nodes = [
            helper.make_node("MatMul", ["input_0", "input_1"], ["product"], "MatMul_0"),
            helper.make_node("Cast", ["product"], ["product_cast"], "Cast_0", to=type1),
            helper.make_node("Cast", ["product_cast"], ["output"], "Cast_1", to=type2),
        ]
        input_type = (
            type2 if type1 != type2 else (TensorProto.FLOAT16 if type1 == TensorProto.FLOAT else TensorProto.FLOAT)
        )
        output_type = (
            input_type
            if type1 != type2
            else (TensorProto.FLOAT16 if input_type == TensorProto.FLOAT else TensorProto.FLOAT)
        )
        inputs = [
            helper.make_tensor_value_info("input_0", input_type, ["M", "K"]),
            helper.make_tensor_value_info("input_1", input_type, ["K", "N"]),
        ]

        outputs = [
            helper.make_tensor_value_info("output", output_type, ["M", "N"]),
        ]

        save(
            model_path + "_" + type_to_string(type1) + "_" + type_to_string(type2),
            nodes,
            inputs,
            outputs,
            [],
        )


def gen_fuse_sibling_casts(model_path):
    for type1, type2 in list(itertools.product([TensorProto.FLOAT, TensorProto.FLOAT16], repeat=2)):
        input_type = (
            type2 if type1 != type2 else (TensorProto.FLOAT16 if type1 == TensorProto.FLOAT else TensorProto.FLOAT)
        )
        nodes = [
            helper.make_node("MatMul", ["input_0", "input_1"], ["product"], "MatMul_0"),
            helper.make_node("Cast", ["product"], ["cast_0_output"], "Cast_0", to=type1),
            helper.make_node("Identity", ["cast_0_output"], ["output_0"], "Identity_0"),
            helper.make_node("Cast", ["product"], ["cast_1_output"], "Cast_1", to=type2),
            helper.make_node("Identity", ["cast_1_output"], ["output_1"], "Identity_1"),
        ]

        inputs = [
            helper.make_tensor_value_info("input_0", input_type, ["M", "K"]),
            helper.make_tensor_value_info("input_1", input_type, ["K", "N"]),
        ]

        outputs = [
            helper.make_tensor_value_info("output_0", type1, ["M", "N"]),
            helper.make_tensor_value_info("output_1", type2, ["M", "N"]),
        ]

        save(
            model_path + "_" + type_to_string(type1) + "_" + type_to_string(type2),
            nodes,
            inputs,
            outputs,
            [],
        )


def flip_type(type, flip=True):
    return (TensorProto.FLOAT16 if type == TensorProto.FLOAT else TensorProto.FLOAT) if flip else type


def do_cast_inputs(input_0, input_1, nodes, input_cast_type):
    nodes.extend(
        [
            helper.make_node("Cast", [input_0], ["cast_" + input_0], "Cast_0", to=input_cast_type),
            helper.make_node("Cast", [input_1], ["cast_" + input_1], "Cast_1", to=input_cast_type),
        ]
    )
    return "cast_" + input_0, "cast_" + input_1


def do_transpose_inputs(input_0, input_1, nodes):
    nodes.extend(
        [
            helper.make_node("Transpose", [input_0], ["input_transpose_0"], "Transpose_0"),
            helper.make_node("Transpose", [input_1], ["input_transpose_1"], "Transpose_1"),
        ]
    )
    return "input_transpose_0", "input_transpose_1"


def do_cast_product(product, nodes, product_type):
    nodes.insert(
        1,
        helper.make_node("Cast", [product], [product + "_cast"], "Cast_2", to=product_type),
    )
    return product + "_cast"


def do_transpose_product(product, nodes):
    if transpose_product:
        nodes.append(helper.make_node("Transpose", [product], [product + "_transpose"], "Transpose_2"))
    return product + "_transpose"


def do_cast_sum(sum, nodes, type):
    nodes.append(helper.make_node("Cast", [sum], ["cast_" + sum], "Cast_3", to=type))
    return "cast_" + sum


def do_cast_input2(input_2, nodes, type):
    nodes.append(helper.make_node("Cast", [input_2], ["cast_" + input_2], "Cast_4", to=type))
    return "cast_" + input_2


def gen_propagate_cast_test_model(
    model_path,
    transpose_inputs,
    transpose_product,
    cast_inputs,
    cast_product,
    insert_add,
    cast_sum,
    cast_input2,
    transpose_inputs_before_cast=False,
):
    input_0 = "input_0"
    input_1 = "input_1"
    product = "product"
    nodes = []
    input_type = TensorProto.FLOAT
    product_type = input_type
    if transpose_inputs_before_cast:
        if transpose_inputs:
            input_0, input_1 = do_transpose_inputs(input_0, input_1, nodes)
        if cast_inputs:
            input_0, input_1 = do_cast_inputs(input_0, input_1, nodes, input_type)
            input_type = flip_type(input_type)
    else:
        if cast_inputs:
            input_0, input_1 = do_cast_inputs(input_0, input_1, nodes, input_type)
            input_type = flip_type(input_type)
        if transpose_inputs:
            input_0, input_1 = do_transpose_inputs(input_0, input_1, nodes)
    nodes.append(helper.make_node("MatMul", [input_0, input_1], [product], "MatMul_0"))
    if transpose_product:
        product = do_transpose_product(product, nodes)

    if cast_product:
        product = do_cast_product(product, nodes, flip_type(product_type))
        product_type = flip_type(product_type)

    inputs = [
        helper.make_tensor_value_info("input_0", input_type, ["N", "N"]),
        helper.make_tensor_value_info("input_1", input_type, ["N", "N"]),
    ]
    if insert_add:
        input_2 = "input_2"
        add_input_type = flip_type(product_type, cast_input2)
        inputs.append(helper.make_tensor_value_info(input_2, add_input_type, ["N", "N"]))
        output = "sum"
        output_type = product_type
        if cast_input2:
            input_2 = do_cast_input2(input_2, nodes, flip_type(add_input_type))
        nodes.append(helper.make_node("Add", [product, input_2], [output], "Add_0"))
        if cast_sum:
            output = do_cast_sum(output, nodes, flip_type(output_type))
            output_type = flip_type(output_type)
    else:
        output = product
        output_type = product_type
    outputs = [helper.make_tensor_value_info(output, output_type, ["N", "N"])]

    save(
        model_path
        + ("_transpose_inputs" if transpose_inputs else "")
        + ("_transpose_product" if transpose_product else "")
        + ("_cast_inputs" if cast_inputs else "")
        + ("_cast_product" if cast_product else "")
        + ("_cast_input2" if cast_input2 else "")
        + ("_cast_sum" if cast_sum else ""),
        nodes,
        inputs,
        outputs,
        [],
    )


def gen_matmul_two_products(model_path, transpose, transpose_before_cast, second_matmul, cast_inputs):
    def do_transpose(output_0, output_1, transpose, nodes):
        nodes.append(helper.make_node("Transpose", [output_0], ["transpose_0_" + output_0], "Transpose_0"))
        output_0 = "transpose_0_" + output_0
        if transpose > 1:
            nodes.append(helper.make_node("Transpose", [output_1], ["transpose_1_" + output_1], "Transpose_1"))
            output_1 = "transpose_1_" + output_1
        return output_0, output_1

    input_type = flip_type(TensorProto.FLOAT, cast_inputs)
    input_0 = "input_0"
    input_1 = "input_1"
    output = "product"
    output_0 = "product"
    output_1 = "product"
    outputs = []
    nodes = []
    cast_count = 0
    inputs = [
        helper.make_tensor_value_info("input_0", input_type, ["M", "K"]),
        helper.make_tensor_value_info("input_1", input_type, ["K", "N"]),
    ]
    if cast_inputs:
        input_type = flip_type(input_type)
        input_0, input_1 = do_cast_inputs(input_0, input_1, nodes, input_type)
        cast_count += 2
    output0_type = input_type
    output1_type = input_type
    nodes.append(helper.make_node("MatMul", [input_0, input_1], [output], "MatMul_0"))
    if second_matmul:
        nodes.append(helper.make_node("MatMul", [input_0, input_1], ["second_" + output], "MatMul_1"))
        outputs.append(helper.make_tensor_value_info("second_" + output, input_type, ["M", "N"]))
        if add_products:
            nodes.append(helper.make_node("Add", [output, "second_" + output], ["sum"], "Add_0"))
            outputs.append(helper.make_tensor_value_info("sum", input_type, ["M", "N"]))
    if transpose > 0 and transpose_before_cast:
        output_0, output_1 = do_transpose(output_0, output_1, transpose, nodes)
    output0_type = flip_type(output0_type)
    nodes.append(
        helper.make_node(
            "Cast",
            [output_0],
            ["cast_" + str(cast_count) + "_" + output_0],
            "Cast_" + str(cast_count),
            to=output0_type,
        )
    )
    output_0 = "cast_" + str(cast_count) + "_" + output_0
    cast_count += 1
    if second_matmul:
        nodes.append(
            helper.make_node(
                "Cast",
                [output_1],
                ["cast_" + str(cast_count) + "_" + output_1],
                "Cast_" + str(cast_count),
                to=TensorProto.FLOAT16,
            )
        )
        output_1 = "cast_" + str(cast_count) + "_" + output_1
        output1_type = flip_type(output1_type)
    if transpose > 0 and not transpose_before_cast:
        output_0, output_1 = do_transpose(output_0, output_1, transpose, nodes)

    outputs.extend(
        [
            helper.make_tensor_value_info(output_0, output0_type, ["M", "N"]),
            helper.make_tensor_value_info(output_1, output1_type, ["M", "N"]),
        ]
    )
    model_path += "_cast_inputs" if cast_inputs else ""
    model_path += (
        ("_transpose_before_cast" if transpose_before_cast else "_transpose_after_cast") if transpose > 0 else ""
    )
    model_path += "_transpose" if transpose > 1 else ""
    model_path += "_second_matmul" if second_matmul else ""
    model_path += "_add_products" if add_products else ""
    save(model_path, nodes, inputs, outputs, [])


def gen_bool_to_float16_cast(model_path):
    X1 = helper.make_tensor_value_info("x1", TensorProto.INT64, [1, 1])  # noqa: N806
    X2 = helper.make_tensor_value_info("x2", TensorProto.INT64, [1, 1])  # noqa: N806
    X3 = helper.make_tensor_value_info("x3", TensorProto.FLOAT, [1, 1])  # noqa: N806
    Y = helper.make_tensor_value_info("output", TensorProto.FLOAT16, [1, 1])  # noqa: N806

    less1 = helper.make_node("Less", ["x1", "x2"], ["less1"], name="less1")
    cast1 = helper.make_node("Cast", ["less1"], ["cast1"], name="cast1", to=TensorProto.FLOAT16)
    cast2 = helper.make_node("Cast", ["x3"], ["cast2"], name="cast2", to=TensorProto.FLOAT16)
    add1 = helper.make_node("Add", ["cast1", "cast2"], ["output"])

    save(model_path, [less1, cast1, cast2, add1], [X1, X2, X3], [Y], [])


def gen_bool_to_float_cast(model_path):
    X1 = helper.make_tensor_value_info("x1", TensorProto.INT64, [1, 1])  # noqa: N806
    X2 = helper.make_tensor_value_info("x2", TensorProto.INT64, [1, 1])  # noqa: N806
    X3 = helper.make_tensor_value_info("x3", TensorProto.FLOAT16, [1, 1])  # noqa: N806
    Y = helper.make_tensor_value_info("output", TensorProto.FLOAT16, [1, 1])  # noqa: N806

    less1 = helper.make_node("Less", ["x1", "x2"], ["less1"], name="less1")
    cast1 = helper.make_node("Cast", ["less1"], ["cast1"], name="cast1", to=TensorProto.FLOAT)
    cast2 = helper.make_node("Cast", ["x3"], ["cast2"], name="cast2", to=TensorProto.FLOAT)
    add1 = helper.make_node("Add", ["cast1", "cast2"], ["add1"])
    cast3 = helper.make_node("Cast", ["add1"], ["output"], name="cast3", to=TensorProto.FLOAT16)

    save(model_path, [less1, cast1, cast2, cast3, add1], [X1, X2, X3], [Y], [])


def gen_one_input_one_output_test(op, model_path, axes_attribute=False):
    X = helper.make_tensor_value_info("x", TensorProto.FLOAT16, [2, 2])  # noqa: N806
    output_shape = [2, 2]
    if op == "Unsqueeze":
        output_shape.append(1)
    Y = helper.make_tensor_value_info("y", TensorProto.FLOAT16, output_shape)  # noqa: N806
    node_inputs = []
    graph_inputs = [X]
    cast1 = helper.make_node("Cast", ["x"], ["cast1"], name="cast1", to=TensorProto.FLOAT)
    node_inputs.insert(0, "cast1")
    if axes_attribute:
        node = helper.make_node(
            op,
            node_inputs,
            ["op_output"],
            name=op + str(1),
            axes=np.array([2]).astype(np.int64),
        )
    else:
        node = helper.make_node(op, node_inputs, ["op_output"], name=op + str(1))
    cast2 = helper.make_node("Cast", ["op_output"], ["y"], name="cast2", to=TensorProto.FLOAT16)
    save(model_path, [cast1, node, cast2], graph_inputs, [Y], [])


for (
    transpose_inputs,
    transpose_product,
    cast_inputs,
    cast_product,
    insert_add,
    cast_sum,
    cast_input2,
) in list(itertools.product([False, True], repeat=7)):
    if not insert_add and (cast_sum or cast_input2):
        continue
    if cast_inputs or cast_product or cast_sum:
        gen_propagate_cast_test_model(
            "matmul_add" if insert_add else "matmul",
            transpose_inputs,
            transpose_product,
            cast_inputs,
            cast_product,
            insert_add,
            cast_sum,
            cast_input2,
        )

gen_fuse_sibling_casts("fuse_sibling_casts")
gen_fuse_back2back_casts("fuse_back2back_casts")

for (
    transpose,
    transpose_before_cast,
    second_matmul,
    add_products,
    cast_inputs,
) in list(itertools.product([0, 1, 2], [False, True], [False, True], [False, True], [False, True])):
    if not transpose and transpose_before_cast:
        continue
    if not second_matmul and add_products:
        continue
    gen_matmul_two_products(
        "matmul_two_outputs",
        transpose,
        transpose_before_cast,
        second_matmul,
        cast_inputs,
    )


gen_bool_to_float16_cast("negative_test_case_bool_fp16_cast")
gen_bool_to_float_cast("negative_test_case_bool_fp_cast")
gen_one_input_one_output_test("Squeeze", "squeeze_cast_propagation_test")
gen_one_input_one_output_test("Unsqueeze", "unsqueeze_cast_propagation_test", True)



import onnx
from onnx import TensorProto, helper


def GenerateModel(model_name, sign, b_zp=True, bias=False):  # noqa: N802
    nodes = [  # DynamicQuantizeMatMul subgraph
        helper.make_node(
            "DynamicQuantizeLinear",
            ["A"],
            ["a_quantized", "a_scale", "a_zp"],
            "DynamicQuantizeLinear",
        ),
        helper.make_node(
            "MatMulInteger",
            ["a_quantized", "B", "a_zp", "b_zero_point"] if b_zp else ["a_quantized", "B", "a_zp"],
            ["matmul_output_int32"],
            "MatMulInteger",
        ),
        helper.make_node("Mul", ["a_scale", "b_scale"], ["multiplier"], "mul_right"),
        helper.make_node("Cast", ["matmul_output_int32"], ["matmul_output_float"], "cast", to=1),
        helper.make_node(
            "Mul",
            ["matmul_output_float", "multiplier"],
            ["mul_bottom_output" if bias else "Y"],
            "mul_bottom",
        ),
    ]

    inputs = [
        helper.make_tensor_value_info("A", TensorProto.FLOAT, ["M", "K"]),
        helper.make_tensor_value_info("B", TensorProto.INT8 if sign else TensorProto.UINT8, ["K", "N"]),
        helper.make_tensor_value_info("b_scale", TensorProto.FLOAT, ["C"]),
    ]

    if b_zp:
        inputs.extend(
            [
                helper.make_tensor_value_info(
                    "b_zero_point",
                    TensorProto.INT8 if sign else TensorProto.UINT8,
                    ["C"],
                )
            ]
        )

    if bias:
        nodes.extend([helper.make_node("Add", ["mul_bottom_output", "bias"], ["Y"], "add")])

        inputs.extend([helper.make_tensor_value_info("bias", TensorProto.FLOAT, ["N"])])

    graph = helper.make_graph(
        nodes,
        "DynamicQuantizeMatMul_fusion",  # name
        inputs,
        [  # outputs
            helper.make_tensor_value_info("Y", TensorProto.FLOAT, ["M", "N"]),
        ],
    )

    model = helper.make_model(graph)
    onnx.save(model, model_name)


if __name__ == "__main__":
    GenerateModel("dynamic_quantize_matmul_int8.onnx", True)
    GenerateModel("dynamic_quantize_matmul_uint8.onnx", False)
    GenerateModel("dynamic_quantize_matmul_int8_bias.onnx", True, False, True)
    GenerateModel("dynamic_quantize_matmul_uint8_bias.onnx", False, False, True)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import logging
from abc import abstractmethod

import onnx

import onnxruntime.training.onnxblock._graph_utils as _graph_utils
import onnxruntime.training.onnxblock._training_graph_utils as _training_graph_utils
import onnxruntime.training.onnxblock.blocks as blocks
import onnxruntime.training.onnxblock.model_accessor as accessor


class ForwardBlock(blocks.Block):
    """Base class for all blocks that require forward model to be automatically built.

    Blocks wanting to build a forward model by stacking blocks on top of the existing model
    must subclass this class. The subclass's implementation of the build method must return the
    name of the graph output. This block will automatically register the output as a graph output
    and build the model.

    Example:

    >>> class MyForwardBlock(ForwardBlock):
    >>>     def __init__(self):
    >>>         super().__init__()
    >>>         self.loss = onnxblock.loss.CrossEntropyLoss()
    >>>
    >>>     def build(self, loss_input_name: str):
    >>>         # Add a cross entropy loss on top of the output so far (loss_input_name)
    >>>         return self.loss(loss_input_name)

    The above example will automatically build the forward graph that is composed of the existing model
    and the cross entropy loss function stacked on top of it.
    """

    def __init__(self):
        super().__init__()
        self._model = None

    @abstractmethod
    def build(self, *args, **kwargs):
        """Customize the forward graph for this model by stacking up blocks on top of the inputs to this function.

        This method should be overridden by the subclass. The output of this method should
        be the name of the graph output.
        """

    def to_model_proto(self):
        """Returns the forward model.

        Returns:
            model (onnx.ModelProto): The forward model.

        Raises:
            RuntimeError: If the build method has not been invoked (i.e. the forward model has not been built yet).
        """
        if self._model is None:
            raise RuntimeError("Please build the forward models first before trying to retrieve it.")
        return self._model

    def __call__(self, *args, **kwargs):
        """Calls the subclass's build method and builds the forward model."""

        self.base = accessor._GLOBAL_ACCESSOR.model

        logging.debug("Building forward block %s", self.__class__.__name__)

        output = self.build(*args, **kwargs)

        self._model = self.infer_shapes_on_base()

        _graph_utils.register_graph_outputs(self._model, output)

        accessor._GLOBAL_ACCESSOR.model.CopyFrom(self._model)

        return output


class TrainingBlock(blocks.Block):
    """Base class for all blocks that require gradient model to be automatically built.

    Blocks that require the gradient graph to be computed based on the output of the block
    must subclass this class. The subclass's implementation of the build method must return
    the name of the output from where backpropagation must begin (typically the name of the
    output from the loss function).

    Example:

    >>> class MyTrainingBlock(TrainingBlock):
    >>>     def __init__(self):
    >>>         super().__init__()
    >>>         self.loss = onnxblock.loss.CrossEntropyLoss()
    >>>
    >>>     def build(self, loss_input_name: str):
    >>>         # Add a cross entropy loss on top of the output so far (loss_input_name)
    >>>         return self.loss(loss_input_name)

    The above example will automatically build the gradient graph for the entire model
    starting from the output of the loss function.
    """

    def __init__(self):
        super().__init__()
        self._requires_grad = set()
        self._frozen_params = set()
        self._parameters = None
        self._training_model = None
        self._eval_model = None

    @abstractmethod
    def build(self, *args, **kwargs):
        """Customize the forward graph for this model by stacking up blocks on top of the inputs to this function.

        This method should be overridden by the subclass. The output of this method should
        be the name of the output from where backpropagation must begin (typically the name
        of the output from the loss function).
        """

    def requires_grad(self, argument_name: str, value: bool = True):
        """Specify whether the argument requires gradient or not.

        The auto-diff will compute the gradient graph for only the arguments that
        require gradient. By default, none of the arguments require gradient.
        The user must explicitly specify which arguments require gradient.

        Args:
            argument_name (str): The name of the argument that require/does not require gradient.
            value (bool): True if the argument requires gradient, False otherwise.
        """
        if value is True:
            self._requires_grad.add(argument_name)
            if argument_name in self._frozen_params:
                self._frozen_params.remove(argument_name)
        else:
            if argument_name in self._requires_grad:
                self._requires_grad.remove(argument_name)
            self._frozen_params.add(argument_name)

    def parameters(self) -> tuple[list[onnx.TensorProto], list[onnx.TensorProto]]:
        """Trainable as well as non-trainable (frozen) parameters of the model.

        Model parameters that are extracted while building the training model
        are returned by this method.

        Note that the parameters are not known before the training model is
        built. As a result, if this method is invoked before the training model
        is built, an exception will be raised.

        Returns:
            trainable_params (list of onnx.TensorProto): The trainable parameters of the model.
            frozen_params (list of onnx.TensorProto): The non-trainable parameters of the model.

        Raises:
            RuntimeError: If the build method has not been invoked (i.e. the training model has not been built yet).
        """
        if self._parameters is None:
            raise RuntimeError("Please build the training model first before trying to retrieve the parameters.")

        return self._parameters

    def to_model_proto(self) -> tuple[onnx.ModelProto, onnx.ModelProto]:
        """Returns the training and eval models.

        Once the gradient graph is built, the training and eval models can be retrieved
        by invoking this method.

        Returns:
            training_model (onnx.ModelProto): The training model.
            eval_model (onnx.ModelProto): The eval model.

        Raises:
            RuntimeError: If the build method has not been invoked (i.e. the training model has not been built yet).
        """
        if self._training_model is None or self._eval_model is None:
            raise RuntimeError("Please build the training and eval models first before trying to retrieve them.")
        return self._training_model, self._eval_model

    def __call__(self, *args, **kwargs):
        """Calls the subclass's build method and builds the gradient graph on top."""

        self.base = accessor._GLOBAL_ACCESSOR.model

        logging.debug("Building training block %s", self.__class__.__name__)

        output = self.build(*args, **kwargs)

        model = self.infer_shapes_on_base()

        _graph_utils.register_graph_outputs(model, output)

        logging.debug("Building gradient graph for training block %s", self.__class__.__name__)

        self._parameters = _training_graph_utils.get_model_parameters(model, self._requires_grad, self._frozen_params)

        # Build the gradient graph. The gradient graph building is composed of the following steps:
        #   - Move all model parameters to model inputs.
        #   - Run orttraining graph transformers on the model.
        #   - Add the gradient graph to the optimized model.
        # The order of model inputs after gradient graph building is: user inputs, model parameters as inputs
        # The order of the model outputs is: user outputs, model parameter gradients (in the order of parameter inputs)
        self._training_model, self._eval_model = _training_graph_utils.build_gradient_graph(
            model, self._requires_grad, self._frozen_params, output, accessor._GLOBAL_CUSTOM_OP_LIBRARY
        )

        logging.debug("Adding gradient accumulation nodes for training block %s", self.__class__.__name__)

        _training_graph_utils.build_gradient_accumulation_graph(self._training_model, self._requires_grad)

        accessor._GLOBAL_ACCESSOR.model.CopyFrom(self._training_model)

        return output



import argparse

import deepspeed
import torch
from deepspeed.pipe import PipelineModule
from torch import nn

from onnxruntime.training.ortmodule import ORTModule

# USAGE:
# pip install deepspeed
# deepspeed orttraining_test_ortmodule_deepspeed_pipeline_parallel.py --deepspeed_config=orttraining_test_ortmodule_deepspeed_pipeline_parallel_config.json --pipeline-parallel-size 2 --steps=100
# expected output : steps: 100 loss: 0.0585 iter time (s): 0.186 samples/sec: 53.694


class SampleData(torch.utils.data.Dataset):
    def __init__(self, x, y):
        self.x = x
        self.y = y

    def __len__(self):
        return x.size()[0]

    def __getitem__(self, idx):
        return self.x[idx], self.y[idx]


def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--local_rank", type=int, default=-1, help="local rank passed from distributed launcher")
    parser.add_argument("-s", "--steps", type=int, default=100, help="quit after this many steps")
    parser.add_argument("-p", "--pipeline-parallel-size", type=int, default=2, help="pipeline parallelism")
    parser.add_argument("--backend", type=str, default="nccl", help="distributed backend")
    parser.add_argument("--seed", type=int, default=0, help="PRNG seed")
    parser.add_argument("--fp16", type=bool, default=False, help="fp16 run")
    parser.add_argument("--run_without_ort", type=bool, default=False, help="onlydeepspeed run")

    parser = deepspeed.add_config_arguments(parser)
    args = parser.parse_args()
    return args


n = 10
d_in = 4
d_hidden = 8
d_out = 3
args = get_args()
torch.cuda.set_device(args.local_rank)
device = torch.device("cuda", args.local_rank)

if args.run_without_ort:
    print("Running deepspeed pipeline parallel module without ORTModule")
else:
    print("Running deepspeed pipeline parallel module with ORTModule")


# dist.init_process_group(backend=args.backend)
deepspeed.init_distributed(dist_backend=args.backend)
torch.manual_seed(args.seed)
# Model.

if args.run_without_ort:
    model = nn.Sequential(
        nn.Linear(d_in, d_hidden),  # Stage 1
        nn.ReLU(),  # Stage 1
        nn.Linear(d_hidden, d_hidden),  # Stage 1
        nn.ReLU(),  # Stage 1
        nn.Linear(d_hidden, d_hidden),  # Stage 2
        nn.ReLU(),  # Stage 2
        nn.Linear(d_hidden, d_out),  # Stage 2
    )


else:
    model = nn.Sequential(
        ORTModule(nn.Linear(d_in, d_hidden).to(device)),  # Stage 1
        nn.ReLU().to(
            device
        ),  # ORTModule(nn.ReLU().to(device)), Stage 1, TODO: ORTModule can wrap Relu once stateless model is supported.
        ORTModule(nn.Linear(d_hidden, d_hidden).to(device)),  # Stage 1
        nn.ReLU().to(
            device
        ),  # ORTModule(nn.ReLU().to(device)), Stage 1, TODO: ORTModule can wrap Relu once stateless model is supported.
        ORTModule(nn.Linear(d_hidden, d_hidden).to(device)),  # Stage 2
        nn.ReLU().to(
            device
        ),  # ORTModule(nn.ReLU().to(device)), Stage 2, TODO: ORTModule can wrap Relu once stateless model is supported.
        ORTModule(nn.Linear(d_hidden, d_out).to(device)),  # Stage 2
    )

model = PipelineModule(
    layers=model,
    loss_fn=torch.nn.CrossEntropyLoss(),
    num_stages=args.pipeline_parallel_size,
    partition_method="uniform",  #'parameters',
    activation_checkpoint_interval=0,
)

params = [p for p in model.parameters() if p.requires_grad]

# Input.
x = torch.rand((n, d_in))
if args.fp16:
    x = x.half()
# Output.
y = torch.randint(0, d_out, (n,))
ds = SampleData(x, y)

print("Initialize deepspeed")
model_engine, optimizer, _, _ = deepspeed.initialize(
    args=args,
    model=model,
    model_parameters=params,
    training_data=ds,  # (x,y)#
)

for step in range(args.steps):
    loss = model_engine.train_batch()
    if step % 10 == 0:
        print("step = ", step, ", loss = ", loss)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------

import onnx
from onnx import OperatorSetIdProto, TensorProto, helper

input_shape = [2, 2]

input = helper.make_tensor_value_info("input", TensorProto.FLOAT, input_shape)
output = helper.make_tensor_value_info("output", TensorProto.FLOAT, input_shape)

cast1 = helper.make_node("Cast", ["input"], ["cast1"], name="cast1", to=2)
shape1 = helper.make_node("Shape", ["cast1"], ["shape1"], name="shape1")
add = helper.make_node("Add", ["cast1", "cast1"], ["add"], name="add")
cast2 = helper.make_node("Cast", ["add"], ["cast2"], name="cast2", to=1)
shape2 = helper.make_node("Shape", ["cast2"], ["shape2"], name="shape2")
shape3 = helper.make_node("Shape", ["cast2"], ["shape3"], name="shape3")
add_shapes = helper.make_node("Add", ["shape2", "shape3"], ["add_shapes"], name="add_shapes")
cast3 = helper.make_node("Cast", ["add_shapes"], ["output"], name="cast3", to=1)

graph_def = helper.make_graph(
    [cast1, shape1, add, cast2, shape2, shape3, add_shapes, cast3],
    "pre_shape_node_elimination.onnx",
    [input],
    [output],
)

opsets = []
onnxdomain = OperatorSetIdProto()
onnxdomain.version = 15
onnxdomain.domain = ""
opsets.append(onnxdomain)

kwargs = {}
kwargs["opset_imports"] = opsets

model_def = helper.make_model(graph_def, producer_name="onnx-example", **kwargs)
onnx.save(model_def, "pre_shape_node_elimination.onnx")



import onnx
from onnx import TensorProto, helper

# CoreML EP currently handles a special case for supporting ArgMax followed by a Cast to int32.
# Please see <repo_root>/onnxruntime/core/providers/coreml/builders/impl/argmax_op_builder.cc and
# <repo_root>/onnxruntime/core/providers/coreml/builders/impl/cast_op_builder.cc.
# This script generates graphs for these cases:
# - An ArgMax followed by a supported Cast to int32 type
# - An ArgMax followed by an unsupported Cast to a type other than int32


def GenerateModel(model_name, cast_to_dtype):  # noqa: N802
    nodes = [
        helper.make_node("ArgMax", ["X"], ["argmax_output_int64"], "argmax", axis=1, keepdims=1),
        helper.make_node("Cast", ["argmax_output_int64"], ["Y"], "cast", to=cast_to_dtype),
    ]

    graph = helper.make_graph(
        nodes,
        "CoreML_ArgMax_Cast_Test",
        [  # input
            helper.make_tensor_value_info("X", TensorProto.FLOAT, [3, 2, 2]),
        ],
        [  # output
            helper.make_tensor_value_info("Y", cast_to_dtype, [3, 1, 2]),
        ],
    )

    model = helper.make_model(graph)
    onnx.save(model, model_name)


if __name__ == "__main__":
    GenerateModel("coreml_argmax_cast_test.onnx", TensorProto.INT32)
    GenerateModel("coreml_argmax_unsupported_cast_test.onnx", TensorProto.UINT32)



#!/usr/bin/env python3
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import argparse
import logging
import os
import sys

from _test_commons import run_subprocess

logging.basicConfig(format="%(asctime)s %(name)s [%(levelname)s] - %(message)s", level=logging.DEBUG)
log = logging.getLogger("ORTModuleTests")


def parse_arguments():
    parser = argparse.ArgumentParser()
    parser.add_argument("--cwd", help="Path to the current working directory")
    parser.add_argument("--mnist", help="Path to the mnist data directory", type=str, default=None)
    parser.add_argument("--bert_data", help="Path to the bert data directory", type=str, default=None)
    parser.add_argument(
        "--transformers_cache", help="Path to the transformers model cache directory", type=str, default=None
    )
    return parser.parse_args()


def get_env_with_transformers_cache(transformers_cache):
    return {"TRANSFORMERS_CACHE": transformers_cache} if transformers_cache else {}


def run_ortmodule_api_tests(cwd, log, transformers_cache):
    log.debug("Running: ORTModule-API tests")

    env = get_env_with_transformers_cache(transformers_cache)

    command = [sys.executable, "-m", "pytest", "-sv", "orttraining_test_ortmodule_api.py"]

    run_subprocess(command, cwd=cwd, log=log, env=env).check_returncode()


def run_ortmodule_ops_tests(cwd, log, transformers_cache):
    log.debug("Running: ORTModule-OPS tests")

    env = get_env_with_transformers_cache(transformers_cache)

    command = [sys.executable, "-m", "pytest", "-sv", "orttraining_test_ortmodule_onnx_ops.py"]

    run_subprocess(command, cwd=cwd, log=log, env=env).check_returncode()


def run_ortmodule_fallback_tests(cwd, log, transformers_cache):
    log.debug("Running: ORTModule-API tests")

    env = get_env_with_transformers_cache(transformers_cache)

    command = [sys.executable, "-m", "pytest", "-sv", "orttraining_test_ortmodule_fallback.py"]

    run_subprocess(command, cwd=cwd, log=log, env=env).check_returncode()


def run_ortmodule_poc_net(cwd, log, no_cuda, data_dir):
    log.debug(f"Running: ORTModule POCNet for MNIST with --no-cuda arg {no_cuda}.")

    command = [sys.executable, "orttraining_test_ortmodule_poc.py"]
    if no_cuda:
        command.extend(["--no-cuda", "--epochs", str(3)])

    if data_dir:
        command.extend(["--data-dir", data_dir])

    run_subprocess(command, cwd=cwd, log=log).check_returncode()


def run_ortmodule_torch_lightning(cwd, log, data_dir):
    log.debug("Running: ORTModule PyTorch Lightning sample .")

    command = [
        sys.executable,
        "orttraining_test_ortmodule_torch_lightning_basic.py",
        "--train-steps=470",
        "--epochs=2",
        "--batch-size=256",
    ]

    if data_dir:
        command.extend(["--data-dir", data_dir])

    run_subprocess(command, cwd=cwd, log=log).check_returncode()


def run_ortmodule_hf_bert_for_sequence_classification_from_pretrained(cwd, log, no_cuda, data_dir, transformers_cache):
    log.debug(f"Running: ORTModule HuggingFace BERT for sequence classification with --no-cuda arg {no_cuda}.")

    env = get_env_with_transformers_cache(transformers_cache)

    command = [sys.executable, "orttraining_test_ortmodule_bert_classifier.py"]
    if no_cuda:
        command.extend(["--no-cuda", "--epochs", str(3)])

    if data_dir:
        command.extend(["--data-dir", data_dir])

    run_subprocess(command, cwd=cwd, log=log, env=env).check_returncode()


def run_ortmodule_custom_autograd_tests(cwd, log):
    log.debug("Running: ORTModule-Custom AutoGrad Functions tests")

    command = [sys.executable, "-m", "pytest", "-sv", "orttraining_test_ortmodule_autograd.py"]

    run_subprocess(command, cwd=cwd, log=log).check_returncode()


def run_ortmodule_hierarchical_ortmodule_tests(cwd, log):
    log.debug("Running: ORTModule-Hierarchical model tests")

    command = [sys.executable, "-m", "pytest", "-sv", "orttraining_test_hierarchical_ortmodule.py"]

    run_subprocess(command, cwd=cwd, log=log).check_returncode()


def run_ortmodule_experimental_json_config_tests(cwd, log):
    log.debug("Running: ORTModule Experimental Load Config tests")

    command = [sys.executable, "-m", "pytest", "-sv", "orttraining_test_ortmodule_experimental_json_config.py"]

    run_subprocess(command, cwd=cwd, log=log).check_returncode()


def run_experimental_gradient_graph_tests(cwd, log):
    log.debug("Running: Experimental Gradient Graph Export Tests")

    command = [sys.executable, "-m", "pytest", "-sv", "orttraining_test_experimental_gradient_graph.py"]

    run_subprocess(command, cwd=cwd, log=log).check_returncode()


def run_data_sampler_tests(cwd, log):
    log.debug("Running: Data sampler tests")

    command = [sys.executable, "-m", "pytest", "-sv", "orttraining_test_sampler.py"]

    run_subprocess(command, cwd=cwd, log=log).check_returncode()


def run_hooks_tests(cwd, log):
    log.debug("Running: Data hooks tests")

    command = [sys.executable, "-m", "pytest", "-sv", "orttraining_test_ortmodule_hooks.py"]

    run_subprocess(command, cwd=cwd, log=log).check_returncode()


def run_utils_tests(cwd, log):
    log.debug("Running: Utils tests")

    command = [sys.executable, "-m", "pytest", "-sv", "orttraining_test_utilities.py"]

    run_subprocess(command, cwd=cwd, log=log).check_returncode()


def run_pytorch_export_contrib_ops_tests(cwd, log):
    log.debug("Running: PyTorch Export Contrib Ops Tests")

    command = [sys.executable, "-m", "pytest", "-sv", "test_pytorch_export_contrib_ops.py"]

    run_subprocess(command, cwd=cwd, log=log).check_returncode()


def main():
    args = parse_arguments()
    cwd = args.cwd

    log.info("Running ortmodule tests pipeline")

    run_ortmodule_api_tests(cwd, log, transformers_cache=args.transformers_cache)

    run_ortmodule_ops_tests(cwd, log, transformers_cache=args.transformers_cache)

    run_ortmodule_poc_net(cwd, log, no_cuda=False, data_dir=args.mnist)

    if os.getenv("ORTMODULE_DISABLE_CPU_TRAINING_TEST", "0") != "1":
        run_ortmodule_poc_net(cwd, log, no_cuda=True, data_dir=args.mnist)

    run_ortmodule_hf_bert_for_sequence_classification_from_pretrained(
        cwd, log, no_cuda=False, data_dir=args.bert_data, transformers_cache=args.transformers_cache
    )

    if os.getenv("ORTMODULE_DISABLE_CPU_TRAINING_TEST", "0") != "1":
        run_ortmodule_hf_bert_for_sequence_classification_from_pretrained(
            cwd, log, no_cuda=True, data_dir=args.bert_data, transformers_cache=args.transformers_cache
        )

    run_ortmodule_torch_lightning(cwd, log, args.mnist)

    run_ortmodule_custom_autograd_tests(cwd, log)

    run_ortmodule_experimental_json_config_tests(cwd, log)

    run_ortmodule_fallback_tests(cwd, log, args.transformers_cache)

    run_ortmodule_hierarchical_ortmodule_tests(cwd, log)

    run_data_sampler_tests(cwd, log)

    run_hooks_tests(cwd, log)

    run_utils_tests(cwd, log)

    run_experimental_gradient_graph_tests(cwd, log)

    # TODO(bmeswani): Enable this test once it can run with latest pytorch
    # run_pytorch_export_contrib_ops_tests(cwd, log)

    return 0


if __name__ == "__main__":
    sys.exit(main())



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------

import os
import shutil
import warnings
from io import TextIOWrapper
from pathlib import Path

import onnx
import torch

from ._subscriber_base import RuntimeStates, SubscriberBase
from ._subscriber_manager import ORT_NO_INCREASE_GLOBAL_STEP


class _InspectActivation(torch.autograd.Function):
    """
    This class is used to run the subscriber's forward and backward functions.
    The function will be called by two kinds of callers:
        1. SubscriberManager calls it for each registered nn.Module.
        2. Users who want to inspect the activation tensor at any place of model definition code.
    """

    @staticmethod
    def forward(
        ctx,
        activation_name: str,
        module_idx: int | None,
        run_ctx: RuntimeStates,
        input_tensor: torch.Tensor,
        module_post_forward,
        module_pre_backward,
    ):
        """
        Args:
            ctx: context object to store intermediate information.
            activation_name: the name of the activation tensor.
            module_idx: unit id of the module (address of the module instance).
            run_ctx: runtime context.
                For call case 2 - need retrieve the runtime state from GlobalSubscriberManager.
            input_tensor: the activation tensor.

        Make sure there is a same number of `tensor` type inputs and outputs.
        This is enforced by ORT's PythonOp's schema check.
        """
        depth = -1
        if module_idx is not None:
            depth = run_ctx.global_states.module_index_to_depth[module_idx]

        input_tensor_copied = None
        if input_tensor is None or not isinstance(input_tensor, torch.Tensor):
            input_tensor_copied = input_tensor
        else:
            input_tensor_copied = input_tensor.detach().clone()

        ctx.current_step = run_ctx.global_states.execution_step
        ctx.name = activation_name
        ctx.id = module_idx
        ctx.depth = depth
        ctx.module_pre_backward = module_pre_backward

        module_post_forward(input_tensor_copied, depth, activation_name, ctx.current_step)

        return input_tensor.detach() if input_tensor is not None else None

    @staticmethod
    def backward(ctx, grad_output: torch.Tensor):
        val = None
        if grad_output is None or not isinstance(grad_output, torch.Tensor):
            val = grad_output
        else:
            val = grad_output.detach().clone()

        ctx.module_pre_backward(val, ctx.depth, ctx.name, ctx.current_step)

        return (
            None,
            None,
            None,
            grad_output.detach() if grad_output is not None else None,
            None,
            None,
        )

    @staticmethod
    def infer_shape(
        node: onnx.NodeProto,
        tensor_input_shapes: list[list[int | str] | None],
        tensor_input_dtypes: list[torch.onnx.TensorProtoDataType],
    ) -> tuple[list[list[int | str] | None], list[torch.onnx.TensorProtoDataType]]:
        return tensor_input_shapes, tensor_input_dtypes

    @staticmethod
    def alias_input(node_proto_str: str):
        fw_alias_map = [3]
        bw_alias_map = [-1] * 6
        bw_alias_map[3] = 0
        return fw_alias_map, bw_alias_map


class StatisticsSubscriber(SubscriberBase):
    """
    This subscriber is used to dump the activation statistics into files.

    Each activation will be summarized into 1 or 2 files, depending on whether it is used in the backward pass.
    > In the forward pass, summarize the tensor's statistics and write to a file.
    > In the backward pass, summarize the tensor's gradient statistics and write it into another file.
    So for each run step, there will be many files.

    Currently, the statistics mainly include:
    > Number of inf/nan values.
    > Common statistics: tensor shape, data type, total element size, min/max/mean/std of the tensor elements.
    > Number of zero/negative/positive elements.
    > Sampled elements (taking the first 128 elements for example).
    > To be extended...

    `merge_activation_summary.py` can be used to merge the files into one file per training step.
    """

    def __init__(
        self,
        output_dir: str,
        start_step: None | int = None,
        end_step: None | int = None,
        override_output_dir: bool = False,
        run_on_cpu: bool = False,
        bucket_size: int = 1024 * 1024 * 1024 // 2,
    ):
        """
        Steps in [start_step, end_step) will run subscriber actions.

        Args:
            output_dir: the directory in all activation statistics files will be stored.
            start_step: the first step that runs subscriber actions.
            end_step: the end step (exclusively) that runs subscriber actions.
            override_output_dir: whether `output_dir` can be overridden if it already exists.
            run_on_cpu: whether to run the subscriber actions on CPU, this should be the last resort when inserted
                inspector node affects memory peak causing the original recipe run to fail with OOM.
            bucket_size: the size of the bucket to split the statistic calculation.
        """
        super().__init__(start_step=start_step, end_step=end_step)
        self._output_dir = output_dir
        self._run_on_cpu = run_on_cpu
        self._bucket_size = bucket_size
        if os.path.exists(self._output_dir):
            if override_output_dir:
                warnings.warn(f"Output directory {self._output_dir} already exists, overriding it.")
                shutil.rmtree(self._output_dir)
            else:
                raise FileExistsError(
                    f"Output directory {self._output_dir} already exists. "
                    "Set override_output_dir=True for StatisticsSubscriber if this is the intention."
                )

    def post_forward_tensor_apply_impl(
        self, run_rtx: RuntimeStates, module: torch.nn.Module, tensor_index: int, tensor: torch.Tensor
    ) -> torch.Tensor:
        module_index = run_rtx.global_states.module_to_module_index[module]
        name = f"{module.__class__.__name__}_{module_index}_{tensor_index}th_output"
        return _InspectActivation.apply(
            name, module_index, run_rtx, tensor, self.module_post_forward_impl, self.module_pre_backward_impl
        )

    def module_post_forward_impl(self, activation: torch.Tensor, depth: int, name: str, step: int):
        output_file_path = os.path.join(f"{self._output_dir}", f"step_{step}")
        return self._summarize_activations(activation, depth, name, output_file_path, True)

    def module_pre_backward_impl(self, activation: torch.Tensor, depth: int, name: str, step: int):
        output_file_path = os.path.join(f"{self._output_dir}", f"step_{step}")
        return self._summarize_activations(activation, depth, name, output_file_path, False)

    def _summarize_activations(self, tensor: torch.Tensor, depth: int, name: str, step_folder: str, is_forward: bool):
        display_name = name + " forward run" if is_forward is True else name + " backward run"
        output_file_name = name + "_forward" if is_forward is True else name + "_backward"

        # Skip dump during model pre-export output schema preparison run and export run.
        if ORT_NO_INCREASE_GLOBAL_STEP[0] is False:
            if tensor is None or not isinstance(tensor, torch.Tensor):
                print(f"{display_name} not a torch tensor, value: {tensor}")
                return

            step_path = Path(step_folder)
            if not step_path.exists():
                step_path.mkdir(parents=True, exist_ok=False)
            order_file_path = step_path / "order.txt"
            tensor_file_path = step_path / output_file_name

            with order_file_path.open(mode="a", encoding="utf-8") as f:
                f.write(f"{output_file_name}\n")

            with tensor_file_path.open(mode="w", encoding="utf-8") as f:
                _summarize_tensor(display_name, tensor, f, depth, self._run_on_cpu, self._bucket_size)


def _summarize_tensor(
    display_name: str,
    tensor: torch.Tensor,
    f: TextIOWrapper,
    depth: int = 0,
    run_on_cpu: bool = False,
    bucket_size: int = 1024 * 1024 * 1024 // 2,
):
    # This is to try the best effort to align the count of numbers per line for easier comparison in diff views,
    # though it does not always guarantee to do this way.
    torch.set_printoptions(precision=6, linewidth=128)

    tensor_shape = tensor.shape
    tensor_dtype = tensor.dtype
    flatten_array = tensor.flatten().view(-1)

    if run_on_cpu:
        flatten_array = flatten_array.to("cpu")

    if run_on_cpu:
        num_nan = torch.isnan(flatten_array).sum()
        num_inf = torch.isinf(flatten_array).sum()
        num_neg = (flatten_array < 0).sum()
        num_pos = (flatten_array > 0).sum()
        num_zero = (flatten_array == 0).sum()
        min_value = flatten_array.min()
        max_value = flatten_array.max()
        mean_value = flatten_array.mean()
        std_value = flatten_array.std()
    else:
        # Split the calculation for each bucket, then do another round of calculation on the bucket results.
        # This can at the best effort reduce the peak memory impact.
        element_count = flatten_array.numel()
        ceil_bucket_count = (element_count + bucket_size - 1) // (bucket_size)
        nan_buckets = torch.zeros(ceil_bucket_count, dtype=torch.int64, device=flatten_array.device)
        inf_buckets = torch.zeros(ceil_bucket_count, dtype=torch.int64, device=flatten_array.device)
        neg_buckets = torch.zeros(ceil_bucket_count, dtype=torch.int64, device=flatten_array.device)
        pos_buckets = torch.zeros(ceil_bucket_count, dtype=torch.int64, device=flatten_array.device)
        zero_buckets = torch.zeros(ceil_bucket_count, dtype=torch.int64, device=flatten_array.device)
        min_buckets = torch.zeros(ceil_bucket_count, dtype=flatten_array.dtype, device=flatten_array.device)
        max_buckets = torch.zeros(ceil_bucket_count, dtype=flatten_array.dtype, device=flatten_array.device)
        mean_buckets = torch.zeros(ceil_bucket_count, dtype=flatten_array.dtype, device=flatten_array.device)
        std_buckets = torch.zeros(ceil_bucket_count, dtype=flatten_array.dtype, device=flatten_array.device)

        # Summary for each bucket
        element_count_per_bucket = torch.zeros(ceil_bucket_count, dtype=torch.int64, device=flatten_array.device)
        for i in range(ceil_bucket_count):
            end = min((i + 1) * bucket_size, element_count)
            bucket = flatten_array[i * bucket_size : end]
            element_count_per_bucket[i] = bucket.numel()

            nan_buckets[i] = torch.isnan(bucket).sum()
            inf_buckets[i] = torch.isinf(bucket).sum()
            neg_buckets[i] = (bucket < 0).sum()
            pos_buckets[i] = (bucket > 0).sum()
            zero_buckets[i] = (bucket == 0).sum()
            min_buckets[i] = bucket.min()
            max_buckets[i] = bucket.max()
            mean_buckets[i] = bucket.sum()

            # Only calculate std for float types, otherwise it will throw exception.
            if bucket.dtype in [torch.float16, torch.float32, torch.float64]:
                std_buckets[i] = bucket.std()

        # Reduction across all buckets
        num_nan = nan_buckets.sum()
        num_inf = inf_buckets.sum()
        num_neg = neg_buckets.sum()
        num_pos = pos_buckets.sum()
        num_zero = zero_buckets.sum()
        min_value = min_buckets.min()
        max_value = max_buckets.max()
        mean_value = float(mean_buckets.sum()) / float(element_count)
        # Here we refer to
        # https://math.stackexchange.com/questions/2971315/how-do-i-combine-standard-deviations-of-two-groups
        # to calculate the combined standard deviation of all buckets.
        s = (element_count_per_bucket - 1) * (std_buckets**2) + element_count_per_bucket * (
            (mean_buckets - mean_value) ** 2
        )
        std_value = torch.sqrt(s.sum() / (element_count - 1))

    f.write(
        f"{'>' * max(0, depth) + display_name} shape: {tensor_shape} dtype: {tensor_dtype} size: {flatten_array.size()} \n"
        f"min: {min_value} max: {max_value}, mean: {mean_value}, "
        f"std: {std_value} \n"
        f"nan: {num_nan}, inf: {num_inf}\n"
    )
    f.write(f"samples(top 128): {flatten_array[:128]}\n")
    f.write(f"neg: {num_neg}, pos: {num_pos}, zero: {num_zero},\n")
    f.write(f"{'=' * 16}\n")



import google.protobuf.text_format
import onnx
from numpy import array, float16

import onnxruntime as ort

# Run n times
N = 1

onnx_model_text = """
ir_version: 8
producer_name: "pytorch"
producer_version: "2.2.0"
graph {
  node {
    output: "_val_1"
    name: "Constant_0"
    op_type: "Constant"
    attribute {
      name: "value_ints"
      ints: -1
      type: INTS
    }
    doc_string: ""
  }
  node {
    input: "input_0"
    input: "_val_1"
    output: "_val_2"
    name: "Reshape_1"
    op_type: "Reshape"
    attribute {
      name: "allowzero"
      i: 0
      type: INT
    }
    doc_string: ""
  }
  node {
    input: "_val_2"
    output: "_val_3"
    name: "_aten_linalg_vector_norm_no_dim_onnx_2"
    op_type: "_aten_linalg_vector_norm_no_dim_onnx"
    attribute {
      name: "keepdim"
      i: 0
      type: INT
    }
    attribute {
      name: "ord"
      f: 2.0
      type: FLOAT
    }
    doc_string: ""
    domain: "pkg.onnxscript.torch_lib"
  }
  name: "main_graph"
  input {
    name: "input_0"
    type {
      tensor_type {
        elem_type: 10
        shape {
        }
      }
    }
  }
  output {
    name: "_val_3"
    type {
      tensor_type {
        elem_type: 10
        shape {
        }
      }
    }
  }
  value_info {
    name: "_val_1"
    type {
      tensor_type {
        elem_type: 7
        shape {
          dim {
            dim_value: 1
          }
        }
      }
    }
  }
  value_info {
    name: "_val_2"
    type {
      tensor_type {
        elem_type: 10
        shape {
          dim {
            dim_value: 1
          }
        }
      }
    }
  }
}
opset_import {
  domain: "pkg.onnxscript.torch_lib"
  version: 1
}
opset_import {
  domain: ""
  version: 18
}
opset_import {
  domain: "pkg.onnxscript.torch_lib.common"
  version: 1
}
functions {
  name: "_aten_linalg_vector_norm_no_dim_onnx"
  input: "self"
  output: "result_29"
  attribute: "ord"
  attribute: "keepdim"
  node {
    input: "self"
    output: "tmp"
    name: "n0"
    op_type: "Shape"
    domain: ""
  }
  node {
    input: "tmp"
    output: "self_rank"
    name: "n1"
    op_type: "Size"
    domain: ""
  }
  node {
    output: "int64_0"
    name: "n2"
    op_type: "Constant"
    attribute {
      name: "value"
      t {
        data_type: 7
        int64_data: 0
        name: "int64_0"
      }
      type: TENSOR
    }
    domain: ""
  }
  node {
    input: "int64_0"
    input: "self_rank"
    output: "int64_0_cast"
    name: "n3"
    op_type: "CastLike"
    domain: ""
  }
  node {
    input: "self_rank"
    input: "int64_0_cast"
    output: "cond"
    name: "n4"
    op_type: "Equal"
    domain: ""
  }
  node {
    input: "cond"
    output: "self_2"
    name: "n5"
    op_type: "If"
    attribute {
      name: "then_branch"
      g {
        node {
          output: "int64_0_1d"
          name: "n0"
          op_type: "Constant"
          attribute {
            name: "value"
            t {
              dims: 1
              data_type: 7
              int64_data: 0
              name: "int64_0_1d"
            }
            type: TENSOR
          }
          domain: ""
        }
        node {
          input: "self"
          input: "int64_0_1d"
          output: "self_0"
          name: "n1"
          op_type: "Unsqueeze"
          domain: ""
        }
        name: "thenGraph_4"
        output {
          name: "self_0"
          type {
          }
        }
      }
      type: GRAPH
    }
    attribute {
      name: "else_branch"
      g {
        node {
          input: "self"
          output: "self_1"
          name: "n0"
          op_type: "Identity"
          domain: ""
        }
        name: "elseGraph_4"
        output {
          name: "self_1"
          type {
          }
        }
      }
      type: GRAPH
    }
    domain: ""
  }
  node {
    input: "self_2"
    output: "self_3"
    name: "n6"
    op_type: "Abs"
    domain: ""
  }
  node {
    output: "ord"
    name: "n7"
    op_type: "Constant"
    attribute {
      name: "value_float"
      type: FLOAT
      ref_attr_name: "ord"
    }
    domain: ""
  }
  node {
    input: "ord"
    output: "ord_4"
    name: "n8"
    op_type: "Cast"
    attribute {
      name: "to"
      i: 1
      type: INT
    }
    domain: ""
  }
  node {
    input: "ord_4"
    output: "cond_5"
    name: "n9"
    op_type: "IsInf"
    attribute {
      name: "detect_negative"
      i: 0
      type: INT
    }
    attribute {
      name: "detect_positive"
      i: 1
      type: INT
    }
    domain: ""
  }
  node {
    input: "cond_5"
    output: "result_24"
    name: "n10"
    op_type: "If"
    attribute {
      name: "then_branch"
      g {
        node {
          input: "self_3"
          output: "result"
          name: "n0"
          op_type: "ReduceMax"
          attribute {
            name: "keepdims"
            type: INT
            ref_attr_name: "keepdim"
          }
          domain: ""
        }
        name: "thenGraph_9"
        output {
          name: "result"
          type {
          }
        }
      }
      type: GRAPH
    }
    attribute {
      name: "else_branch"
      g {
        node {
          input: "ord_4"
          output: "cond_6"
          name: "n0"
          op_type: "IsInf"
          attribute {
            name: "detect_negative"
            i: 1
            type: INT
          }
          attribute {
            name: "detect_positive"
            i: 0
            type: INT
          }
          domain: ""
        }
        node {
          input: "cond_6"
          output: "result_23"
          name: "n1"
          op_type: "If"
          attribute {
            name: "then_branch"
            g {
              node {
                input: "self_3"
                output: "result_7"
                name: "n0"
                op_type: "ReduceMin"
                attribute {
                  name: "keepdims"
                  type: INT
                  ref_attr_name: "keepdim"
                }
                domain: ""
              }
              name: "thenGraph_11"
              output {
                name: "result_7"
                type {
                }
              }
            }
            type: GRAPH
          }
          attribute {
            name: "else_branch"
            g {
              node {
                output: "const"
                name: "n0"
                op_type: "Constant"
                attribute {
                  name: "value"
                  t {
                    data_type: 1
                    float_data: 0.0
                    name: "const"
                  }
                  type: TENSOR
                }
                domain: ""
              }
              node {
                input: "const"
                input: "ord_4"
                output: "const_cast"
                name: "n1"
                op_type: "CastLike"
                domain: ""
              }
              node {
                input: "ord_4"
                input: "const_cast"
                output: "cond_8"
                name: "n2"
                op_type: "Equal"
                domain: ""
              }
              node {
                input: "cond_8"
                output: "result_22"
                name: "n3"
                op_type: "If"
                attribute {
                  name: "then_branch"
                  g {
                    node {
                      input: "self_3"
                      output: "self_bool"
                      name: "n0"
                      op_type: "Cast"
                      attribute {
                        name: "to"
                        i: 9
                        type: INT
                      }
                      domain: ""
                    }
                    node {
                      input: "self_bool"
                      input: "self_3"
                      output: "self_0_1"
                      name: "n1"
                      op_type: "CastLike"
                      domain: ""
                    }
                    node {
                      input: "self_0_1"
                      output: "result_9"
                      name: "n2"
                      op_type: "ReduceSum"
                      attribute {
                        name: "keepdims"
                        i: 0
                        type: INT
                      }
                      domain: ""
                    }
                    name: "thenGraph_13"
                    output {
                      name: "result_9"
                      type {
                      }
                    }
                  }
                  type: GRAPH
                }
                attribute {
                  name: "else_branch"
                  g {
                    node {
                      output: "const_10"
                      name: "n0"
                      op_type: "Constant"
                      attribute {
                        name: "value"
                        t {
                          data_type: 1
                          float_data: 1.0
                          name: "const_10"
                        }
                        type: TENSOR
                      }
                      domain: ""
                    }
                    node {
                      input: "const_10"
                      input: "ord_4"
                      output: "const_10_cast"
                      name: "n1"
                      op_type: "CastLike"
                      domain: ""
                    }
                    node {
                      input: "ord_4"
                      input: "const_10_cast"
                      output: "cond_11"
                      name: "n2"
                      op_type: "Equal"
                      domain: ""
                    }
                    node {
                      input: "cond_11"
                      output: "result_21"
                      name: "n3"
                      op_type: "If"
                      attribute {
                        name: "then_branch"
                        g {
                          node {
                            input: "self_3"
                            output: "result_12"
                            name: "n0"
                            op_type: "ReduceL1"
                            attribute {
                              name: "keepdims"
                              type: INT
                              ref_attr_name: "keepdim"
                            }
                            domain: ""
                          }
                          name: "thenGraph_18"
                          output {
                            name: "result_12"
                            type {
                            }
                          }
                        }
                        type: GRAPH
                      }
                      attribute {
                        name: "else_branch"
                        g {
                          node {
                            output: "const_13"
                            name: "n0"
                            op_type: "Constant"
                            attribute {
                              name: "value"
                              t {
                                data_type: 1
                                float_data: 2.0
                                name: "const_13"
                              }
                              type: TENSOR
                            }
                            domain: ""
                          }
                          node {
                            input: "const_13"
                            input: "ord_4"
                            output: "const_13_cast"
                            name: "n1"
                            op_type: "CastLike"
                            domain: ""
                          }
                          node {
                            input: "ord_4"
                            input: "const_13_cast"
                            output: "cond_14"
                            name: "n2"
                            op_type: "Equal"
                            domain: ""
                          }
                          node {
                            input: "cond_14"
                            output: "result_20"
                            name: "n3"
                            op_type: "If"
                            attribute {
                              name: "then_branch"
                              g {
                                node {
                                  input: "self_3"
                                  output: "result_15"
                                  name: "n0"
                                  op_type: "ReduceL2"
                                  attribute {
                                    name: "keepdims"
                                    type: INT
                                    ref_attr_name: "keepdim"
                                  }
                                  domain: ""
                                }
                                name: "thenGraph_20"
                                output {
                                  name: "result_15"
                                  type {
                                  }
                                }
                              }
                              type: GRAPH
                            }
                            attribute {
                              name: "else_branch"
                              g {
                                node {
                                  input: "ord_4"
                                  input: "self_3"
                                  output: "ord_float"
                                  name: "n0"
                                  op_type: "CastLike"
                                  domain: ""
                                }
                                node {
                                  input: "self_3"
                                  input: "ord_float"
                                  output: "self_pow"
                                  name: "n1"
                                  op_type: "Pow"
                                  domain: ""
                                }
                                node {
                                  input: "self_pow"
                                  output: "tmp_16"
                                  name: "n2"
                                  op_type: "ReduceSum"
                                  attribute {
                                    name: "keepdims"
                                    type: INT
                                    ref_attr_name: "keepdim"
                                  }
                                  domain: ""
                                }
                                node {
                                  output: "const_17"
                                  name: "n3"
                                  op_type: "Constant"
                                  attribute {
                                    name: "value"
                                    t {
                                      data_type: 1
                                      float_data: 1.0
                                      name: "const_17"
                                    }
                                    type: TENSOR
                                  }
                                  domain: ""
                                }
                                node {
                                  input: "const_17"
                                  input: "ord_float"
                                  output: "const_17_cast"
                                  name: "n4"
                                  op_type: "CastLike"
                                  domain: ""
                                }
                                node {
                                  input: "const_17_cast"
                                  input: "ord_float"
                                  output: "tmp_18"
                                  name: "n5"
                                  op_type: "Div"
                                  domain: ""
                                }
                                node {
                                  input: "tmp_16"
                                  input: "tmp_18"
                                  output: "result_19"
                                  name: "n6"
                                  op_type: "Pow"
                                  domain: ""
                                }
                                name: "elseGraph_20"
                                output {
                                  name: "result_19"
                                  type {
                                  }
                                }
                              }
                              type: GRAPH
                            }
                            domain: ""
                          }
                          name: "elseGraph_18"
                          output {
                            name: "result_20"
                            type {
                            }
                          }
                        }
                        type: GRAPH
                      }
                      domain: ""
                    }
                    name: "elseGraph_13"
                    output {
                      name: "result_21"
                      type {
                      }
                    }
                  }
                  type: GRAPH
                }
                domain: ""
              }
              name: "elseGraph_11"
              output {
                name: "result_22"
                type {
                }
              }
            }
            type: GRAPH
          }
          domain: ""
        }
        name: "elseGraph_9"
        output {
          name: "result_23"
          type {
          }
        }
      }
      type: GRAPH
    }
    domain: ""
  }
  node {
    output: "int64_0_25"
    name: "n11"
    op_type: "Constant"
    attribute {
      name: "value"
      t {
        data_type: 7
        int64_data: 0
        name: "int64_0_25"
      }
      type: TENSOR
    }
    domain: ""
  }
  node {
    input: "int64_0_25"
    input: "self_rank"
    output: "int64_0_25_cast"
    name: "n12"
    op_type: "CastLike"
    domain: ""
  }
  node {
    input: "self_rank"
    input: "int64_0_25_cast"
    output: "cond_26"
    name: "n13"
    op_type: "Equal"
    domain: ""
  }
  node {
    input: "cond_26"
    output: "result_29"
    name: "n14"
    op_type: "If"
    attribute {
      name: "then_branch"
      g {
        node {
          input: "result_24"
          output: "result_27"
          name: "n0"
          op_type: "Squeeze"
          domain: ""
        }
        name: "thenGraph_27"
        output {
          name: "result_27"
          type {
          }
        }
      }
      type: GRAPH
    }
    attribute {
      name: "else_branch"
      g {
        node {
          input: "result_24"
          output: "result_28"
          name: "n0"
          op_type: "Identity"
          domain: ""
        }
        name: "elseGraph_27"
        output {
          name: "result_28"
          type {
          }
        }
      }
      type: GRAPH
    }
    domain: ""
  }
  opset_import {
    domain: ""
    version: 18
  }
  domain: "pkg.onnxscript.torch_lib"
}
functions {
  name: "Rank"
  input: "input"
  output: "return_val"
  node {
    input: "input"
    output: "tmp"
    name: "n0"
    op_type: "Shape"
    domain: ""
  }
  node {
    input: "tmp"
    output: "return_val"
    name: "n1"
    op_type: "Size"
    domain: ""
  }
  doc_string: "Take the rank of the input tensor."
  opset_import {
    domain: ""
    version: 18
  }
  domain: "pkg.onnxscript.torch_lib.common"
}
functions {
  name: "IsScalar"
  input: "input"
  output: "return_val"
  node {
    input: "input"
    output: "tmp"
    name: "n0"
    op_type: "Shape"
    domain: ""
  }
  node {
    input: "tmp"
    output: "tmp_0"
    name: "n1"
    op_type: "Size"
    domain: ""
  }
  node {
    output: "tmp_1"
    name: "n2"
    op_type: "Constant"
    attribute {
      name: "value_int"
      i: 0
      type: INT
    }
    domain: ""
  }
  node {
    input: "tmp_0"
    input: "tmp_1"
    output: "return_val"
    name: "n3"
    op_type: "Equal"
    domain: ""
  }
  doc_string: "Return whether the input has rank 0, or is a scalar."
  opset_import {
    domain: ""
    version: 18
  }
  domain: "pkg.onnxscript.torch_lib.common"
}

"""

ort_inputs = {"input_0": array(0.8965, dtype=float16)}

# Set up the inference session
session_options = ort.SessionOptions()
session_options.graph_optimization_level = ort.GraphOptimizationLevel.ORT_DISABLE_ALL
onnx_model = onnx.ModelProto()
google.protobuf.text_format.Parse(onnx_model_text, onnx_model)

# Uncomment this line to save the model to a file for examination
# onnx.save_model(onnx_model, "transform_nested_ifs_toplogical_sorted_nodes.onnx")

onnx.checker.check_model(onnx_model)
session = ort.InferenceSession(onnx_model.SerializeToString(), session_options, providers=("CPUExecutionProvider",))

# Run the model
for _ in range(N):
    ort_outputs = session.run(None, ort_inputs)



import argparse
import csv
import os  # noqa: F401

import numpy as np  # noqa: F401

parser = argparse.ArgumentParser()
parser.add_argument("--input", type=str)
args = parser.parse_args()


def get_gpu_lines(path):
    lines = []
    with open(path, newline="") as f:
        reader = csv.reader(f, delimiter=",")
        for row in reader:
            if row[2].find("TotalDurationNs") < 0:
                lines.append(row)
        return lines


activities = [
    ("nccl", lambda x: x.find("nccl") >= 0),
    ("gemm", lambda x: x.find("Cijk_") >= 0),
    ("memcpy", lambda x: x.find("CUDA mem") >= 0),
    ("adam", lambda x: x.lower().find("adam") >= 0),
    ("lamb", lambda x: x.lower().find("lamb") >= 0 or x.lower().find("multi_tensor_apply") >= 0),
    ("dropout", lambda x: x.lower().find("dropout") >= 0 or x.find("curand") >= 0),
    ("layernorm", lambda x: x.find("LayerNorm") >= 0 or x.find("cuCompute") >= 0),
    ("reduce", lambda x: x.find("reduce") >= 0),
    ("softmax", lambda x: x.lower().find("softmax") >= 0),
    ("transpose", lambda x: x.lower().find("transpose") >= 0),
    ("element-wise", lambda x: x.lower().find("elementwise") >= 0 or x.find("DivGrad") >= 0),
    ("jit", lambda x: x.startswith("kernel_")),
    ("misc", lambda x: True),
]


def group_gpu_activity(lines):
    groups = {name: [] for name, _ in activities}
    for line in lines:
        for name, check in activities:
            if check(line[0]):
                groups[name].append(line)
                break
    return groups


def get_seconds(time):
    return float(time.replace("us", "")) / (1000.0 * 1000.0 * 1000.0)


def gpu_percent_time(activities):
    return sum([float(a[4].replace("%", "")) for a in activities])


def gpu_absolute_time(activities):
    return sum([get_seconds(a[2]) for a in activities])


def gpu_kernel_calls(activities):
    return sum([int(a[1]) for a in activities])


lines = get_gpu_lines(args.input)
groups = group_gpu_activity(lines)

for name in groups:
    activities = groups[name]
    print(
        f"{name}: N={len(activities)}, calls={gpu_kernel_calls(activities)}, absolute={gpu_absolute_time(activities):.3f}s, percent={gpu_percent_time(activities):.2f}%"
    )

total = [item for name in groups for item in groups[name]]
print(
    f"Total: N={len(total)}, calls={gpu_kernel_calls(total)}, absolute={gpu_absolute_time(total):.3f}s, percent={gpu_percent_time(total):.2f}%"
)



"""
Github publishes the markdown documentation with jekyll enabled.
This extension does not publish any folder starting with `_`.
These folders need to be renamed.
"""

import os
import re


def rename_folder(root):
    """
    Renames all folder starting with `_`.
    Returns the list of renamed folders.
    """
    found = []
    for r, dirs, _files in os.walk(root):
        for name in dirs:
            if name.startswith("_"):
                found.append((r, name))
    renamed = []
    for r, name in found:
        into = name.lstrip("_")
        renamed.append((r, name, into))
        full_src = os.path.join(r, name)
        full_into = os.path.join(r, into)
        if os.path.exists(full_into):
            raise RuntimeError("%r already exists, previous documentation should be removed.")
        print(f"rename {full_src!r}")
        os.rename(full_src, full_into)

    return renamed


def replace_files(root, renamed):
    subs = {r[1]: r[2] for r in renamed}
    reg = re.compile('(\\"[a-zA-Z0-9\\.\\/\\?\\:@\\-_=#]+\\.([a-zA-Z]){2,6}([a-zA-Z0-9\\.\\&\\/\\?\\:@\\-_=#])*\\")')

    for r, _dirs, files in os.walk(root):
        for name in files:
            if os.path.splitext(name)[-1] != ".html":
                continue
            full = os.path.join(r, name)
            with open(full, encoding="utf-8") as f:
                content = f.read()
            find = reg.findall(content)
            repl = []
            for f in find:
                if f[0].startswith("http"):
                    continue
                for k, v in subs.items():
                    if k == v:
                        raise ValueError(f"{k!r} == {v!r}")
                    if (f'"{k}') in f[0]:
                        repl.append((f[0], f[0].replace(f'"{k}', f'"{v}')))
                    if (f"/{k}") in f[0]:
                        repl.append((f[0], f[0].replace(f"/{k}", f"/{v}")))
            if len(repl) == 0:
                continue
            print(f"update {full!r}")
            for k, v in repl:
                content = content.replace(k, v)
            with open(full, "w", encoding="utf-8") as f:
                f.write(content)


if __name__ == "__main__":
    import sys

    if len(sys.argv) > 1:
        root = sys.argv[-1]
    else:
        root = "../../build/docs/html"
    print(f"look into {root!r}")
    ren = rename_folder(root)
    if len(ren) == 0:
        ren = [
            ("", "_static", "static"),
            ("", "_images", "images"),
            ("", "_downloads", "downloads"),
            ("", "_sources", "sources"),
            ("", "_modules", "modules"),
        ]
    replace_files(root, ren)
    print("done.")



import onnx
from onnx import OperatorSetIdProto, TensorProto, helper

# inputs/outputs
A = helper.make_tensor_value_info("A", TensorProto.FLOAT, ["unk_1", "unk_2", 1024, 4096])
B = helper.make_tensor_value_info("B", TensorProto.FLOAT, ["unk_1", "unk_2", 1024, 4096])

# initializers
quant_scale = helper.make_tensor("quant_scale", TensorProto.FLOAT, [], [0.075])
quant_zero_point_int = helper.make_tensor("quant_zero_point", TensorProto.INT8, [], [20])
quant_zero_point_uint = helper.make_tensor("quant_zero_point", TensorProto.UINT8, [], [120])

# opsets
opsets = []
onnxdomain = OperatorSetIdProto()
onnxdomain.version = 12
onnxdomain.domain = ""
opsets.append(onnxdomain)
kwargs = {}
kwargs["opset_imports"] = opsets

# graph nodes
quantize_linear = helper.make_node(
    "QuantizeLinear", ["A", "quant_scale", "quant_zero_point"], ["quantized_value"], "quanztize0"
)
dequantize_linear = helper.make_node(
    "DequantizeLinear", ["quantized_value", "quant_scale", "quant_zero_point"], ["B"], "dequanztize0"
)
quantize_linear_zp_not_provided = helper.make_node(
    "QuantizeLinear", ["A", "quant_scale"], ["quantized_value"], "quanztize0"
)
dequantize_linear_zp_not_provided = helper.make_node(
    "DequantizeLinear", ["quantized_value", "quant_scale"], ["B"], "dequanztize0"
)

# graph with int8 quantization
graph = helper.make_graph(
    [quantize_linear, dequantize_linear],
    "QDQFusion",
    [A],
    [B],
    [quant_scale, quant_zero_point_int],
)

# model with int8 quantization
model = helper.make_model(graph, producer_name="onnx-example", **kwargs)
onnx.save(model, "qdq_fusion_int8.onnx")

# graph with uint8 quantization
graph = helper.make_graph(
    [quantize_linear, dequantize_linear],
    "QDQFusion",
    [A],
    [B],
    [quant_scale, quant_zero_point_uint],
)

# model with uint8 quantization
model = helper.make_model(graph, producer_name="onnx-example", **kwargs)
onnx.save(model, "qdq_fusion_uint8.onnx")

# graph with no zero point initializer
graph = helper.make_graph(
    [quantize_linear_zp_not_provided, dequantize_linear_zp_not_provided],
    "QDQFusion",
    [A],
    [B],
    [quant_scale],
)

# model with no zero point initializer
model = helper.make_model(graph, producer_name="onnx-example", **kwargs)
onnx.save(model, "qdq_fusion_zp_not_provided.onnx")



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import os
import tempfile

import pytest
import torch

from onnxruntime.training.ortmodule import ORTModule
from onnxruntime.training.utils.hooks import GlobalSubscriberManager, StatisticsSubscriber, inspect_activation


class NeuralNetSingleOutput(torch.nn.Module):
    def __init__(self, input_size, hidden_size, num_classes):
        super().__init__()

        self.fc1 = torch.nn.Linear(input_size, hidden_size)
        self.relu = torch.nn.ReLU()
        self.fc2 = torch.nn.Linear(hidden_size, num_classes)

    def forward(self, input1, input2):
        model_input = input1 + input2
        out = self.fc1(model_input)
        out = self.relu(out)
        out = self.fc2(out)
        return out


class NeuralNetMultipleOutputs(torch.nn.Module):
    def __init__(self, input_size, hidden_size, num_classes):
        super().__init__()

        self.fc1 = torch.nn.Linear(input_size, hidden_size)
        self.relu = torch.nn.ReLU()
        self.fc2 = torch.nn.Linear(hidden_size, num_classes)

    def forward(self, input1, input2):
        model_input = input1 + input2
        out = self.fc1(model_input)
        out = self.relu(out)
        out = self.fc2(out)
        return out, model_input


@pytest.mark.parametrize("device", ["cpu", "cuda"])
@pytest.mark.parametrize("backend", ["torch", "ortmodule"])
def test_statistic_subscriber_single_output(device, backend):
    input_size = 8
    hidden_size = 16
    num_classes = 32
    model = NeuralNetSingleOutput(input_size, hidden_size, num_classes)
    model.to(device)
    model.train()

    with tempfile.TemporaryDirectory() as temporary_dir:
        output_dir_path = os.path.join(temporary_dir, f"{backend}_out")
        GlobalSubscriberManager.subscribe(model, [StatisticsSubscriber(output_dir_path, override_output_dir=True)])

        if backend == "ortmodule":
            model = ORTModule(model)

        batch_size = 4
        input1_tensor = torch.randn(batch_size, input_size, device=device)
        input2_tensor = torch.randn(batch_size, input_size, device=device)
        for _ in range(5):
            y = model(input1_tensor, input2_tensor)
            y.sum().backward()

        assert os.path.exists(output_dir_path)

        expected_files = [
            "order.txt",
            "Linear_1_0th_output_forward",
            "Linear_1_0th_output_backward",
            "NeuralNetSingleOutput_0_0th_output_forward",
            "NeuralNetSingleOutput_0_0th_output_backward",
            "ReLU_2_0th_output_forward",
            "ReLU_2_0th_output_backward",
            "Linear_3_0th_output_forward",
            "Linear_3_0th_output_backward",
        ]

        for i in range(5):
            step_dir = os.path.join(output_dir_path, f"step_{i}")
            for file in expected_files:
                assert os.path.exists(os.path.join(step_dir, file))


@pytest.mark.parametrize("device", ["cpu", "cuda"])
@pytest.mark.parametrize("backend", ["torch", "ortmodule"])
def test_statistic_subscriber_multiple_outputs(device, backend):
    input_size = 8
    hidden_size = 16
    num_classes = 32
    model = NeuralNetMultipleOutputs(input_size, hidden_size, num_classes)
    model.to(device)
    model.train()

    with tempfile.TemporaryDirectory() as temporary_dir:
        output_dir_path = os.path.join(temporary_dir, f"{backend}_out")
        GlobalSubscriberManager.subscribe(model, [StatisticsSubscriber(output_dir_path, override_output_dir=True)])

        if backend == "ortmodule":
            model = ORTModule(model)

        batch_size = 4
        input1_tensor = torch.randn(batch_size, input_size, device=device).requires_grad_(True)
        input2_tensor = torch.randn(batch_size, input_size, device=device)
        for _ in range(5):
            y_output1, y_output2 = model(input1_tensor, input2_tensor)
            y = y_output1.sum() + y_output2.sum()
            y.backward()

        assert os.path.exists(output_dir_path)

        expected_files = [
            "order.txt",
            "NeuralNetMultipleOutputs_0_0th_output_forward",
            "NeuralNetMultipleOutputs_0_0th_output_backward",
            "NeuralNetMultipleOutputs_0_1th_output_forward",
            "NeuralNetMultipleOutputs_0_1th_output_backward",
            "Linear_1_0th_output_forward",
            "Linear_1_0th_output_backward",
            "ReLU_2_0th_output_forward",
            "ReLU_2_0th_output_backward",
            "Linear_3_0th_output_forward",
            "Linear_3_0th_output_backward",
        ]

        for i in range(5):
            step_dir = os.path.join(output_dir_path, f"step_{i}")

            assert len(os.listdir(step_dir)) == len(expected_files)
            for file in expected_files:
                assert os.path.exists(os.path.join(step_dir, file))


class NeuralNetUserAnnotateIntermediateTensor(torch.nn.Module):
    def __init__(self, input_size, hidden_size, num_classes):
        super().__init__()

        self.fc1 = torch.nn.Linear(input_size, hidden_size)
        self.relu = torch.nn.ReLU()
        self.fc2 = torch.nn.Linear(hidden_size, num_classes)

    def forward(self, input1, input2):
        model_input = input1 + input2
        out = self.fc1(model_input)
        out = inspect_activation("fc1_out", out)
        out = self.relu(out)
        out = inspect_activation("relu_out", out)
        out = self.fc2(out)
        return out


@pytest.mark.parametrize("device", ["cpu", "cuda"])
@pytest.mark.parametrize("backend", ["torch", "ortmodule"])
def test_statistic_subscriber_user_annotate_intermediate_tensors(device, backend):
    input_size = 8
    hidden_size = 16
    num_classes = 32
    model = NeuralNetUserAnnotateIntermediateTensor(input_size, hidden_size, num_classes)
    model.to(device)
    model.train()

    with tempfile.TemporaryDirectory() as temporary_dir:
        output_dir_path = os.path.join(temporary_dir, f"{backend}_out")
        GlobalSubscriberManager.subscribe(model, [StatisticsSubscriber(output_dir_path, override_output_dir=True)])

        if backend == "ortmodule":
            model = ORTModule(model)

        batch_size = 4
        input1_tensor = torch.randn(batch_size, input_size, device=device)
        input2_tensor = torch.randn(batch_size, input_size, device=device)
        for _ in range(5):
            y = model(input1_tensor, input2_tensor)
            y.sum().backward()

        assert os.path.exists(output_dir_path)

        expected_files = [
            "order.txt",
            "Linear_1_0th_output_forward",
            "Linear_1_0th_output_backward",
            "NeuralNetUserAnnotateIntermediateTensor_0_0th_output_forward",
            "NeuralNetUserAnnotateIntermediateTensor_0_0th_output_backward",
            "ReLU_2_0th_output_forward",
            "ReLU_2_0th_output_backward",
            "Linear_3_0th_output_forward",
            "Linear_3_0th_output_backward",
            "fc1_out_forward",
            "fc1_out_backward",
            "relu_out_forward",
            "relu_out_backward",
        ]

        for i in range(5):
            step_dir = os.path.join(output_dir_path, f"step_{i}")
            for file in expected_files:
                assert os.path.exists(os.path.join(step_dir, file))
